#
# coding=utf-8
"""test_edist - Test edist interface"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

from doctest import _ellipsis_match as ellipsis_match
from unittest import TestCase

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO  # NOQA

from expecter import expect
from mock import patch

from upoints.edist import (LocationsError, NumberedPoint, NumberedPoints,
                           main, read_csv)


class TestLocationsError(TestCase):
    with expect.raises(LocationsError, 'Invalid location data.'):
        raise LocationsError()
    with expect.raises(LocationsError,
                       'More than one location is required for distance.'):
        raise LocationsError('distance')
    with expect.raises(LocationsError,
                       "Location parsing failure in location 4 '52;None'."):
        raise LocationsError(data=(4, '52;None'))


class TestNumberedPoint(TestCase):
    def test___repr__(self):
        expect(repr(NumberedPoint(52.015, -0.221, 4))) == \
            "NumberedPoint(52.015, -0.221, 4, 'metric')"
        expect(repr(NumberedPoint(52.015, -0.221, 'Home'))) == \
            "NumberedPoint(52.015, -0.221, 'Home', 'metric')"


class TestNumberedPoints(TestCase):
    def test___repr__(self):
        locations = ['0;0'] * 4
        expect(repr(NumberedPoints(locations))) == \
            "NumberedPoints([NumberedPoint(0.0, 0.0, 1, 'metric'), NumberedPoint(0.0, 0.0, 2, 'metric'), NumberedPoint(0.0, 0.0, 3, 'metric'), NumberedPoint(0.0, 0.0, 4, 'metric')], 'dd', True, None, 'km')"

    def test_import_locations(self):
        locs = NumberedPoints(['0;0', 'Home', '0;0'],
                              config_locations={'Home': (52.015, -0.221)})
        expect(repr(locs)) == "NumberedPoints([NumberedPoint(0.0, 0.0, 1, 'metric'), NumberedPoint(52.015, -0.221, 'Home', 'metric'), NumberedPoint(0.0, 0.0, 3, 'metric')], 'dd', True, {'Home': (52.015, -0.221)}, 'km')"

    @patch('sys.stdout', new_callable=StringIO)
    def test_display(self, stdout):
        locs = NumberedPoints(['Home', '52.168;0.040'],
                              config_locations={'Home': (52.015, -0.221)})
        locs.display(None)
        expect(stdout.getvalue()) == (
            "Location Home is 52\xc2\xb000.90'N, 000\xc2\xb013.26'W\n"
            "Location 2 is 52\xc2\xb010.08'N, 000\xc2\xb002.40'E\n"
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_display_locator(self, stdout):
        locs = NumberedPoints(['Home', '52.168;0.040'],
                              config_locations={'Home': (52.015, -0.221)})
        locs.format = 'locator'
        locs.display('extsquare')
        expect(stdout.getvalue()) == (
            'Location Home is IO92va33\n'
            'Location 2 is JO02ae40\n'
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_display_non_verbose(self, stdout):
        locs = NumberedPoints(['Home', '52.168;0.040'],
                              config_locations={'Home': (52.015, -0.221)})
        locs.format = 'locator'
        locs.verbose = False
        locs.display('extsquare')
        expect(stdout.getvalue()) == 'IO92va33\nJO02ae40\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_distance(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.distance()
        expect(stdout.getvalue()) == 'Location 1 to 2 is 24 kilometres\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_distance_sm(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'],
                                   units='sm')
        locations.distance()
        expect(stdout.getvalue()) == 'Location 1 to 2 is 15 miles\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_distance_nm(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'],
                                   units='nm')
        locations.verbose = False
        locations.distance()
        # Manually convert to string here to workaround Python 2/3 float
        # formatting differences
        expect(stdout.getvalue()) == str(13.298957431655218) + '\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_distance_multi(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040',
                                   '51.420;-1.500'])
        locations.distance()
        expect(stdout.getvalue()) == (
            'Location 1 to 2 is 24 kilometres\n'
            'Location 2 to 3 is 134 kilometres\n'
            'Total distance is 159 kilometres\n'
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_bearing(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.bearing('bearing', False)
        expect(stdout.getvalue()) == 'Location 1 to 2 is 46°\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_bearing_symbolic(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.bearing('bearing', True)
        expect(stdout.getvalue()) == 'Location 1 to 2 is North-east\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_final_bearing(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.bearing('final_bearing', False)
        expect(stdout.getvalue()) == \
            'Final bearing from location 1 to 2 is 46°\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_final_bearing_symbolic(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.bearing('final_bearing', True)
        expect(stdout.getvalue()) == \
            'Final bearing from location 1 to 2 is North-east\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_bearing_non_verbose(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.verbose = False
        locations.bearing('bearing', True)
        expect(stdout.getvalue()) == 'North-east\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_final_bearing_non_verbose(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.verbose = False
        locations.bearing('final_bearing', True)
        expect(stdout.getvalue()) == 'North-east\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_range(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.range(20)
        expect(stdout.getvalue()) == \
            'Location 2 is not within 20 kilometres of location 1\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_range2(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.range(30)
        expect(stdout.getvalue()) == \
            'Location 2 is within 30 kilometres of location 1\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_range_non_verbose(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.verbose = False
        locations.range(30)
        expect(stdout.getvalue()) == 'True\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_destination(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.destination(42, 240, False)
        expect(stdout.getvalue()) == (
            "Destination from location 1 is 52\xc2\xb000.90'N, 000\xc2\xb013.26'W\n"
            "Destination from location 2 is 52\xc2\xb010.08'N, 000\xc2\xb002.40'E\n"
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_destination_locator(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.format = 'locator'
        locations.destination(42, 240, 'subsquare')
        expect(stdout.getvalue()) == (
            'Destination from location 1 is IO91ot\n'
            'Destination from location 2 is IO91sx\n'
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_destination_locator_non_verbose(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.format = 'locator'
        locations.verbose = False
        locations.destination(42, 240, 'extsquare')
        expect(stdout.getvalue()) == 'IO91ot97\nIO91sx14\n'

    @patch('sys.stdout', new_callable=StringIO)
    def test_sunrise(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.sun_events('sunrise')
        lines = stdout.getvalue().splitlines()
        expect(ellipsis_match('Sunrise at ... in location 1', lines[0])) \
            == True
        expect(ellipsis_match('Sunrise at ... in location 2', lines[1])) \
            == True

    @patch('sys.stdout', new_callable=StringIO)
    def test_sunset(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040'])
        locations.sun_events('sunset')
        lines = stdout.getvalue().splitlines()
        expect(ellipsis_match('Sunset at ... in location 1', lines[0])) \
            == True
        expect(ellipsis_match('Sunset at ... in location 2', lines[1])) \
            == True

    @patch('sys.stdout', new_callable=StringIO)
    def test_flight_plan(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040',
                                   '52.249;0.130', '52.494;0.654'])
        locations.flight_plan(0, 'h')
        expect(stdout.getvalue()) == (
            'WAYPOINT,BEARING[°],DISTANCE[km],ELAPSED_TIME[h],LATITUDE[d.dd],LONGITUDE[d.dd]\n'
            '1,,,,52.015000,-0.221000\n'
            '2,46,24.6,,52.168000,0.040000\n'
            '3,34,10.9,,52.249000,0.130000\n'
            '4,52,44.8,,52.494000,0.654000\n'
            '-- OVERALL --#,,80.3,,,\n'
            '-- DIRECT --#,47,79.9,,,\n'
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_flight_plan_minute(self, stdout):
        locations = NumberedPoints(['52.015;-0.221', '52.168;0.040',
                                    '52.249;0.130', '52.494;0.654'],
                                   units='nm')
        locations.flight_plan(20, 'm')
        expect(stdout.getvalue()) == (
            'WAYPOINT,BEARING[°],DISTANCE[nm],ELAPSED_TIME[m],LATITUDE[d.dd],LONGITUDE[d.dd]\n'
            '1,,,,52.015000,-0.221000\n'
            '2,46,13.3,0.7,52.168000,0.040000\n'
            '3,34,5.9,0.3,52.249000,0.130000\n'
            '4,52,24.2,1.2,52.494000,0.654000\n'
            '-- OVERALL --,,43.4,2.2,,\n'
            '-- DIRECT --,47,43.1,2.2,,\n'
        )


def test_read_csv():
    locations, names = read_csv(open('tests/data/gpsbabel'))
    expect(sorted(locations.items())) == \
        [('01:My place', ('52.01500', '-0.22100')),
         ('02:Microsoft Research Cambridge', ('52.16700', '00.39000'))]
    expect(names) == ['01:My place', '02:Microsoft Research Cambridge']


@patch('sys.stdout', new_callable=StringIO)
@patch.object(sys, 'argv', ['edist', 'display', '52.015;-0.221'])
def test_main(stdout):
    main()
    expect(stdout.getvalue()) == \
        "Location 1 is 52\xc2\xb000.90'N, 000\xc2\xb013.26'W\n"



#
# coding=utf-8
"""point - Classes for working with locations on Earth"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division

__doc__ += """.

.. moduleauthor:: James Rowe <jnrowe@gmail.com>
.. versionadded:: 0.1.0
"""

import math

from upoints import utils
from upoints.compat import mangle_repr_type


def _manage_location(attr):
    """Build managed property interface.

    :param str attr: Property's name
    :rtype: ``property``
    :return: Managed property interface

    """
    return property(lambda self: getattr(self, '_%s' % attr),
                    lambda self, value: self._set_location(attr, value))


def _dms_formatter(latitude, longitude, mode, unistr=False):
    """Generate a human readable DM/DMS location string.

    :param float latitude: Location's latitude
    :param float longitude: Location's longitude
    :param str mode: Coordinate formatting system to use
    :param bool unistr: Whether to use extended character set

    """
    if unistr:
        chars = ('°', '′', '″')
    else:
        chars = ('°', "'", '"')

    latitude_dms = tuple(map(abs, utils.to_dms(latitude, mode)))
    longitude_dms = tuple(map(abs, utils.to_dms(longitude, mode)))
    text = []
    if mode == 'dms':
        text.append('%%02i%s%%02i%s%%02i%s' % chars % latitude_dms)
    else:
        text.append('%%02i%s%%05.2f%s' % chars[:2] % latitude_dms)
    text.append('S' if latitude < 0 else 'N')
    if mode == 'dms':
        text.append(', %%03i%s%%02i%s%%02i%s' % chars % longitude_dms)
    else:
        text.append(', %%03i%s%%05.2f%s' % chars[:2] % longitude_dms)
    text.append('W' if longitude < 0 else 'E')
    return text


@mangle_repr_type
class Point(object):

    """Simple class for representing a location on a sphere.

    .. versionadded:: 0.2.0

    """

    __slots__ = ('units', '_latitude', '_longitude', '_rad_latitude',
                 '_rad_longitude', 'timezone', '_angle')

    def __init__(self, latitude, longitude, units='metric',
                 angle='degrees', timezone=0):
        """Initialise a new ``Point`` object.

        :type latitude: ``float``, ``tuple`` or ``list``
        :param float latitude: Location's latitude
        :type longitude: ``float``, ``tuple`` or ``list``
        :param float longitude: Location's longitude
        :param str angle: Type for specified angles
        :param str units: Units type to be used for distances
        :param int timezone: Offset from UTC in minutes
        :raise ValueError: Unknown value for ``angle``
        :raise ValueError: Unknown value for ``units``
        :raise ValueError: Invalid value for ``latitude`` or ``longitude``

        """
        super(Point, self).__init__()
        if angle in ('degrees', 'radians'):
            self._angle = angle
        else:
            raise ValueError('Unknown angle type %r' % angle)
        self._set_location('latitude', latitude)
        self._set_location('longitude', longitude)

        if units in ('imperial', 'metric', 'nautical'):
            self.units = units
        elif units == 'km':
            self.units = 'metric'
        elif units in ('US customary', 'sm'):
            self.units = 'imperial'
        elif units == 'nm':
            self.units = 'nautical'
        else:
            raise ValueError('Unknown units type %r' % units)
        self.timezone = timezone

    def _set_location(self, ltype, value):
        """Check supplied location data for validity, and update."""
        if self._angle == 'degrees':
            if isinstance(value, (tuple, list)):
                value = utils.to_dd(*value)
            setattr(self, '_%s' % ltype, float(value))
            setattr(self, '_rad_%s' % ltype, math.radians(float(value)))
        elif self._angle == 'radians':
            setattr(self, '_rad_%s' % ltype, float(value))
            setattr(self, '_%s' % ltype, math.degrees(float(value)))
        else:
            raise ValueError('Unknown angle type %r' % self._angle)
        if ltype == 'latitude' and not -90 <= self._latitude <= 90:
            raise ValueError('Invalid latitude value %r' % value)
        elif ltype == 'longitude' and not -180 <= self._longitude <= 180:
            raise ValueError('Invalid longitude value %r' % value)
    latitude = _manage_location('latitude')
    longitude = _manage_location('longitude')
    rad_latitude = _manage_location('rad_latitude')
    rad_longitude = _manage_location('rad_longitude')

    @property
    def __dict__(self):
        """Emulate ``__dict__`` class attribute for class.

        :rtype: ``dict``
        :return: Object attributes, as would be provided by a class that didn't
            set ``__slots__``

        """
        slots = []
        cls = self.__class__
        # Build a tuple of __slots__ from all parent classes
        while not cls is object:
            slots.extend(cls.__slots__)
            cls = cls.__base__
        return dict((item, getattr(self, item)) for item in slots)

    def __repr__(self):
        """Self-documenting string representation.

        :rtype: ``str``
        :return: String to recreate ``Point`` object

        """
        return utils.repr_assist(self, {'angle': 'degrees'})

    def __str__(self):
        """Pretty printed location string.

        :rtype: ``str``
        :return: Human readable string representation of ``Point`` object

        """
        return format(self)

    def __unicode__(self):
        """Pretty printed Unicode location string.

        :rtype: ``str``
        :return: Human readable Unicode representation of ``Point`` object

        """
        return _dms_formatter(self, 'dd', True)

    def __format__(self, format_spec='dd'):
        """Extended pretty printing for location strings.

        :param str format_spec: Coordinate formatting system to use
        :rtype: ``str``
        :return: Human readable string representation of ``Point`` object
        :raise ValueError: Unknown value for ``format_spec``

        """
        text = []
        if not format_spec:  # default format calls set format_spec to ''
            format_spec = 'dd'
        if format_spec == 'dd':
            text.append('S' if self.latitude < 0 else 'N')
            text.append('%06.3f°; ' % abs(self.latitude))
            text.append('W' if self.longitude < 0 else 'E')
            text.append('%07.3f°' % abs(self.longitude))
        elif format_spec in ('dm', 'dms'):
            text = _dms_formatter(self.latitude, self.longitude, format_spec)
        elif format_spec == 'locator':
            text.append(self.to_grid_locator())
        else:
            raise ValueError('Unknown format_spec %r' % format_spec)

        return ''.join(text)

    def __eq__(self, other, accuracy=None):
        """Compare ``Point`` objects for equality with optional accuracy amount.

        :param Point other: Object to test for equality against
        :param float accuracy: Objects are considered equal if within
            ``accuracy`` ``units`` distance of each other
        :rtype: ``bool``
        :return: True if objects are equal within given bounds

        """
        if accuracy is None:
            return hash(self) == hash(other)
        else:
            return self.distance(other) < accuracy

    def __ne__(self, other, accuracy=None):
        """Compare ``Point`` objects for inequality with optional accuracy amount.

        :param Point other: Object to test for inequality against
        :param float accuracy: Objects are considered equal if within
            ``accuracy`` ``units`` distance
        :rtype: ``bool``
        :return: True if objects are not equal within given bounds

        """
        return not self.__eq__(other, accuracy)

    def __hash__(self):
        """Produce an object hash for equality checks.

        This method returns the hash of the return value from the ``__str__``
        method.  It guarantees equality for objects that have the same latitude
        and longitude.

        .. seealso::

           :meth:`__str__`

        :rtype: ``int``
        :return: Hash of string representation

        """
        return hash(repr(self))

    def to_grid_locator(self, precision='square'):
        """Calculate Maidenhead locator from latitude and longitude.

        :param str precision: Precision with which generate locator string
        :rtype: ``str``
        :return: Maidenhead locator for latitude and longitude

        """
        return utils.to_grid_locator(self.latitude, self.longitude, precision)

    def distance(self, other, method='haversine'):
        """Calculate the distance from self to other.

        As a smoke test this check uses the example from Wikipedia's
        `Great-circle distance entry`_ of Nashville International Airport to
        Los Angeles International Airport, and is correct to within
        2 kilometres of the calculation there.

        :param Point other: Location to calculate distance to
        :param str method: Method used to calculate distance
        :rtype: ``float``
        :return: Distance between self and other in ``units``
        :raise ValueError: Unknown value for ``method``

        .. _Great-circle distance entry:
           http://en.wikipedia.org/wiki/Great-circle_distance

        """
        longitude_difference = other.rad_longitude - self.rad_longitude
        latitude_difference = other.rad_latitude - self.rad_latitude

        if method == 'haversine':
            temp = math.sin(latitude_difference / 2) ** 2 + \
                math.cos(self.rad_latitude) * \
                math.cos(other.rad_latitude) * \
                math.sin(longitude_difference / 2) ** 2
            distance = 2 * utils.BODY_RADIUS * math.atan2(math.sqrt(temp),
                                                          math.sqrt(1 - temp))
        elif method == 'sloc':
            distance = math.acos(math.sin(self.rad_latitude) *
                                 math.sin(other.rad_latitude) +
                                 math.cos(self.rad_latitude) *
                                 math.cos(other.rad_latitude) *
                                 math.cos(longitude_difference)) * \
                utils.BODY_RADIUS
        else:
            raise ValueError('Unknown method type %r' % method)

        if self.units == 'imperial':
            return distance / utils.STATUTE_MILE
        elif self.units == 'nautical':
            return distance / utils.NAUTICAL_MILE
        else:
            return distance

    def bearing(self, other, format='numeric'):
        """Calculate the initial bearing from self to other.

        .. note::
           Applying common plane Euclidean trigonometry to bearing calculations
           suggests to us that the bearing between point A to point B is equal
           to the inverse of the bearing from Point B to Point A, whereas
           spherical trigonometry is much more fun.  If the ``bearing`` method
           doesn't make sense to you when calculating return bearings there are
           plenty of resources on the web that explain spherical geometry.

        .. todo:: Add Rhumb line calculation

        :param Point other: Location to calculate bearing to
        :param str format: Format of the bearing string to return
        :rtype: ``float``
        :return: Initial bearing from self to other in degrees
        :raise ValueError: Unknown value for ``format``

        """
        longitude_difference = other.rad_longitude - self.rad_longitude

        y = math.sin(longitude_difference) * math.cos(other.rad_latitude)
        x = math.cos(self.rad_latitude) * math.sin(other.rad_latitude) - \
            math.sin(self.rad_latitude) * math.cos(other.rad_latitude) * \
            math.cos(longitude_difference)
        bearing = math.degrees(math.atan2(y, x))
        # Always return positive North-aligned bearing
        bearing = (bearing + 360) % 360
        if format == 'numeric':
            return bearing
        elif format == 'string':
            return utils.angle_to_name(bearing)
        else:
            raise ValueError('Unknown format type %r' % format)

    def midpoint(self, other):
        """Calculate the midpoint from self to other.

        .. seealso::

           :meth:`bearing`

        :param Point other: Location to calculate midpoint to
        :rtype: ``Point`` instance
        :return: Great circle midpoint from self to other

        """
        longitude_difference = other.rad_longitude - self.rad_longitude
        y = math.sin(longitude_difference) * math.cos(other.rad_latitude)
        x = math.cos(other.rad_latitude) * math.cos(longitude_difference)
        latitude = math.atan2(math.sin(self.rad_latitude)
                              + math.sin(other.rad_latitude),
                              math.sqrt((math.cos(self.rad_latitude) + x) ** 2
                                        + y ** 2))
        longitude = self.rad_longitude \
            + math.atan2(y, math.cos(self.rad_latitude) + x)

        return Point(latitude, longitude, angle='radians')

    def final_bearing(self, other, format='numeric'):
        """Calculate the final bearing from self to other.

        .. seealso::

           :meth:`bearing`

        :param Point other: Location to calculate final bearing to
        :param str format: Format of the bearing string to return
        :rtype: ``float``
        :return: Final bearing from self to other in degrees
        :raise ValueError: Unknown value for ``format``

        """
        final_bearing = (other.bearing(self) + 180) % 360
        if format == 'numeric':
            return final_bearing
        elif format == 'string':
            return utils.angle_to_name(final_bearing)
        else:
            raise ValueError('Unknown format type %r' % format)

    def destination(self, bearing, distance):
        """Calculate the destination from self given bearing and distance.

        :param float bearing: Bearing from self
        :param float distance: Distance from self in ``self.units``
        :rtype: ``Point``
        :return: Location after travelling ``distance`` along ``bearing``

        """
        bearing = math.radians(bearing)

        if self.units == 'imperial':
            distance *= utils.STATUTE_MILE
        elif self.units == 'nautical':
            distance *= utils.NAUTICAL_MILE

        angular_distance = distance / utils.BODY_RADIUS

        dest_latitude = math.asin(math.sin(self.rad_latitude) *
                                  math.cos(angular_distance) +
                                  math.cos(self.rad_latitude) *
                                  math.sin(angular_distance) *
                                  math.cos(bearing))
        dest_longitude = self.rad_longitude + \
            math.atan2(math.sin(bearing) *
                       math.sin(angular_distance) *
                       math.cos(self.rad_latitude),
                       math.cos(angular_distance) -
                       math.sin(self.rad_latitude) *
                       math.sin(dest_latitude))

        return Point(dest_latitude, dest_longitude, angle='radians')

    def sunrise(self, date=None, zenith=None):
        """Calculate the sunrise time for a ``Point`` object.

        .. seealso::

           :func:`utils.sun_rise_set`

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate rise/set events, or twilight times
        :rtype: :class:`datetime.datetime`
        :return: The time for the given event in the specified timezone

        """
        return utils.sun_rise_set(self.latitude, self.longitude, date, 'rise',
                                  self.timezone, zenith)

    def sunset(self, date=None, zenith=None):
        """Calculate the sunset time for a ``Point`` object.

        .. seealso::

           :func:`utils.sun_rise_set`

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate rise/set events, or twilight times
        :rtype: :class:`datetime.datetime`
        :return: The time for the given event in the specified timezone

        """
        return utils.sun_rise_set(self.latitude, self.longitude, date, 'set',
                                  self.timezone, zenith)

    def sun_events(self, date=None, zenith=None):
        """Calculate the sunrise time for a ``Point`` object.

        .. seealso::

           :func:`utils.sun_rise_set`

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate rise/set events, or twilight times
        :rtype: ``tuple`` of :class:`datetime.datetime`
        :return: The time for the given events in the specified timezone

        """
        return utils.sun_events(self.latitude, self.longitude, date,
                                self.timezone, zenith)

    # Inverse and forward are the common functions expected by people that are
    # familiar with geodesics.
    def inverse(self, other):
        """Calculate the inverse geodesic from self to other.

        :param Point other: Location to calculate inverse geodesic to
        :rtype: ``tuple`` of ``float`` objects
        :return: Bearing and distance from self to other

        """
        return (self.bearing(other), self.distance(other))
    # Forward geodesic function maps directly to destination method
    forward = destination


class TimedPoint(Point):

    """Class for representing a location with an associated time.

    .. versionadded:: 0.12.0

    """

    __slots__ = ('time', )

    def __init__(self, latitude, longitude, units='metric',
                 angle='degrees', timezone=0, time=None):
        """Initialise a new ``TimedPoint`` object.

        :type latitude: ``float``, ``tuple`` or ``list``
        :param latitude: Location's latitude
        :type longitude: ``float``, ``tuple`` or ``list``
        :param longitude: Location's longitude
        :param str angle: Type for specified angles
        :param units: Units type to be used for distances
        :param timezone: Offset from UTC in minutes
        :param datetime.datetime time: Time associated with the location

        """
        super(TimedPoint, self).__init__(latitude, longitude, units, angle,
                                         timezone)
        self.time = time


@mangle_repr_type
class Points(list):

    """Class for representing a group of :class:`Point` objects.

    .. versionadded:: 0.2.0

    """

    def __init__(self, points=None, parse=False, units='metric'):
        """Initialise a new ``Points`` object.

        :type points: ``list`` of `Point` objects
        :param points: :class:`Point` objects to wrap
        :param bool parse: Whether to attempt import of ``points``
        :param str units: Unit type to be used for distances when parsing string
            locations

        """
        super(Points, self).__init__()
        self._parse = parse
        self.units = units
        if points:
            if parse:
                self.import_locations(points)
            else:
                if not all(x for x in points if isinstance(x, Point)):
                    raise TypeError('All `points` elements must be an '
                                    'instance of the `Point` class')
                self.extend(points)

    def __repr__(self):
        """Self-documenting string representation.

        :rtype: ``str``
        :return: String to recreate ``Points`` object

        """
        return utils.repr_assist(self, {'points': self[:]})

    def import_locations(self, locations):
        """Import locations from arguments.

        :type locations: ``list`` of ``str`` or ``tuple``
        :param locations: Location identifiers

        """
        for location in locations:
            data = utils.parse_location(location)
            if data:
                latitude, longitude = data
            else:
                latitude, longitude = utils.from_grid_locator(location)
            self.append(Point(latitude, longitude, self.units))

    def distance(self, method='haversine'):
        """Calculate distances between locations.

        :param str method: Method used to calculate distance
        :rtype: ``list`` of ``float``
        :return: Distance between points in series

        """
        if not len(self) > 1:
            raise RuntimeError('More than one location is required')
        return (self[i].distance(self[i + 1], method)
                for i in range(len(self) - 1))

    def bearing(self, format='numeric'):
        """Calculate bearing between locations.

        :param str format: Format of the bearing string to return
        :rtype: ``list`` of ``float``
        :return: Bearing between points in series

        """
        if not len(self) > 1:
            raise RuntimeError('More than one location is required')
        return (self[i].bearing(self[i + 1], format)
                for i in range(len(self) - 1))

    def final_bearing(self, format='numeric'):
        """Calculate final bearing between locations.

        :param str format: Format of the bearing string to return
        :rtype: ``list`` of ``float``
        :return: Bearing between points in series

        """
        if len(self) == 1:
            raise RuntimeError('More than one location is required')
        return (self[i].final_bearing(self[i + 1], format)
                for i in range(len(self) - 1))

    def inverse(self):
        """Calculate the inverse geodesic between locations.

        :rtype: ``list`` of 2 ``tuple`` of ``float``
        :return: Bearing and distance between points in series

        """
        return ((self[i].bearing(self[i + 1]), self[i].distance(self[i + 1]))
                for i in range(len(self) - 1))

    def midpoint(self):
        """Calculate the midpoint between locations.

        :rtype: ``list`` of :class:`Point` instances
        :return: Midpoint between points in series

        """
        return (self[i].midpoint(self[i + 1]) for i in range(len(self) - 1))

    def range(self, location, distance):
        """Test whether locations are within a given range of ``location``

        :param Point location: Location to test range against
        :param float distance: Distance to test location is within
        :rtype: ``list`` of :class:`Point` objects within specified range
        :return: Points within range of the specified location

        """
        return (x for x in self if location.__eq__(x, distance))

    def destination(self, bearing, distance):
        """Calculate destination locations for given distance and bearings.

        :param float bearing: Bearing to move on in degrees
        :param float distance: Distance in kilometres
        :rtype: ``list`` of :class:`Point`
        :return: Points shifted by ``distance`` and ``bearing``

        """
        return (x.destination(bearing, distance) for x in self)
    forward = destination

    def sunrise(self, date=None, zenith=None):
        """Calculate sunrise times for locations.

        :param datetime.date date: Calculate sunrise for given date
        :param str zenith: Calculate sunrise events, or end of twilight
        :rtype: ``list`` of :class:`datetime.datetime`
        :return: The time for the sunrise for each point

        """
        return (x.sunrise(date, zenith) for x in self)

    def sunset(self, date=None, zenith=None):
        """Calculate sunset times for locations.

        :param datetime.date date: Calculate sunset for given date
        :param str zenith: Calculate sunset events, or start of twilight
        :rtype: ``list`` of :class:`datetime.datetime`
        :return: The time for the sunset for each point

        """
        return (x.sunset(date, zenith) for x in self)

    def sun_events(self, date=None, zenith=None):
        """Calculate sunrise/sunset times for locations.

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate rise/set events, or twilight times
        :rtype: ``list`` of 2 ``tuple`` of :class:`datetime.datetime`
        :return: The time for the sunrise and sunset events for each point

        """
        return (x.sun_events(date, zenith) for x in self)

    def to_grid_locator(self, precision='square'):
        """Calculate Maidenhead locator for locations.

        :param str precision: Precision with which generate locator string
        :rtype: ``list`` of ``str``
        :return: Maidenhead locator for each point

        """
        return (x.to_grid_locator(precision) for x in self)


class TimedPoints(Points):
    def speed(self):
        """Calculate speed between :class:`Points`

        :rtype: ``list`` of ``float``
        :return: Speed between :class:`Point` elements in km/h

        """
        if not len(self) > 1:
            raise RuntimeError('More than one location is required')
        try:
            times = [i.time for i in self]
        except AttributeError:
            raise NotImplementedError('Not all Point objects include time '
                                      'attribute')

        return (distance / ((times[i + 1] - times[i]).seconds / 3600)
                for i, distance in enumerate(self.distance()))


@mangle_repr_type
class KeyedPoints(dict):

    """Class for representing a keyed group of :class:`Point` objects.

    .. versionadded:: 0.2.0

    """

    def __init__(self, points=None, parse=False, units='metric'):
        """Initialise a new ``KeyedPoints`` object.

        :type points: ``dict`` of :class:`Point` objects
        :param points: :class:`Point` objects to wrap
        :param bool points: Whether to attempt import of ``points``
        :param str units: Unit type to be used for distances when parsing string
            locations

        """
        super(KeyedPoints, self).__init__()
        self._parse = parse
        self.units = units
        if points:
            if parse:
                self.import_locations(points)
            else:
                if not all(x for x in points.values() if isinstance(x, Point)):
                    raise TypeError("All `points` element's values must be an "
                                    'instance of the `Point` class')
                self.update(points)

    def __repr__(self):
        """Self-documenting string representation.

        :rtype: ``str``
        :return: String to recreate ``KeyedPoints`` object

        """
        return utils.repr_assist(self, {'points': dict(self.items())})

    def import_locations(self, locations):
        """Import locations from arguments.

        :type locations: ``list`` of 2 ``tuple`` of ``str``
        :param locations: Identifiers and locations

        """
        for identifier, location in locations:
            data = utils.parse_location(location)
            if data:
                latitude, longitude = data
            else:
                latitude, longitude = utils.from_grid_locator(location)
            self[identifier] = Point(latitude, longitude, self.units)

    def distance(self, order, method='haversine'):
        """Calculate distances between locations.

        :param list order: Order to process elements in
        :param str method: Method used to calculate distance
        :rtype: ``list`` of ``float``
        :return: Distance between points in ``order``

        """
        if not len(self) > 1:
            raise RuntimeError('More than one location is required')
        return (self[order[i]].distance(self[order[i + 1]], method)
                for i in range(len(order) - 1))

    def bearing(self, order, format='numeric'):
        """Calculate bearing between locations.

        :param list order: Order to process elements in
        :param str format: Format of the bearing string to return
        :rtype: ``list`` of ``float``
        :return: Bearing between points in series

        """
        if not len(self) > 1:
            raise RuntimeError('More than one location is required')
        return (self[order[i]].bearing(self[order[i + 1]], format)
                for i in range(len(order) - 1))

    def final_bearing(self, order, format='numeric'):
        """Calculate final bearing between locations.

        :param list order: Order to process elements in
        :param str format: Format of the bearing string to return
        :rtype: ``list`` of ``float``
        :return: Bearing between points in series

        """
        if len(self) == 1:
            raise RuntimeError('More than one location is required')
        return (self[order[i]].final_bearing(self[order[i + 1]], format)
                for i in range(len(order) - 1))

    def inverse(self, order):
        """Calculate the inverse geodesic between locations.

        :param list order: Order to process elements in
        :rtype: ``list`` of 2 ``tuple`` of ``float``
        :return: Bearing and distance between points in series

        """
        return ((self[order[i]].bearing(self[order[i + 1]]),
                 self[order[i]].distance(self[order[i + 1]]))
                for i in range(len(order) - 1))

    def midpoint(self, order):
        """Calculate the midpoint between locations.

        :param list order: Order to process elements in
        :rtype: ``list`` of `Point` instance
        :return: Midpoint between points in series

        """
        return (self[order[i]].midpoint(self[order[i + 1]])
                for i in range(len(order) - 1))

    def range(self, location, distance):
        """Test whether locations are within a given range of the first.

        :param Point location: Location to test range against
        :param float distance: Distance to test location is within
        :rtype: ``list`` of :class:`Point` objects within specified range

        """
        return (x for x in self.items() if location.__eq__(x[1], distance))

    def destination(self, bearing, distance):
        """Calculate destination locations for given distance and bearings.

        :param float bearing: Bearing to move on in degrees
        :param float distance: Distance in kilometres

        """
        return ((x[0], x[1].destination(bearing, distance))
                for x in self.items())
    forward = destination

    def sunrise(self, date=None, zenith=None):
        """Calculate sunrise times for locations.

        :param datetime.date date: Calculate sunrise for given date
        :param str zenith: Calculate sunrise events, or end of twilight
        :rtype: ``list`` of :class:`datetime.datetime`
        :return: The time for the sunrise for each point

        """
        return ((x[0], x[1].sunrise(date, zenith)) for x in self.items())

    def sunset(self, date=None, zenith=None):
        """Calculate sunset times for locations.

        :param datetime.date date: Calculate sunset for given date
        :param str zenith: Calculate sunset events, or start of twilight
        :rtype: ``list`` of :class:`datetime.datetime`
        :return: The time for the sunset for each point

        """
        return ((x[0], x[1].sunset(date, zenith)) for x in self.items())

    def sun_events(self, date=None, zenith=None):
        """Calculate sunrise/sunset times for locations.

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate rise/set events, or twilight times
        :rtype: ``list`` of 2 ``tuple`` of :class:`datetime.datetime`
        :return: The time for the sunrise and sunset events for each point

        """
        return ((x[0], x[1].sun_events(date, zenith)) for x in self.items())

    def to_grid_locator(self, precision='square'):
        """Calculate Maidenhead locator for locations.

        :param str precision: Precision with which generate locator string
        :rtype: ``list`` of ``str``
        :return: Maidenhead locator for each point

        """
        return ((x[0], x[1].to_grid_locator(precision)) for x in self.items())



#
# coding=utf-8
"""upoints - Modules for working with points on Earth"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from upoints import _version


__version__ = _version.dotted
__date__ = _version.date
__author__ = 'James Rowe <jnrowe@gmail.com>'
__copyright__ = 'Copyright (C) 2007-2014  James Rowe <jnrowe@gmail.com>'
__license__ = 'GNU General Public License Version 3'
__credits__ = 'Cédric Dufour, Thomas Traber, Kelly Turner, Simon Woods'
__history__ = 'See git repository'

from email.utils import parseaddr

__doc__ += """.

``upoints`` is a collection of `GPL v3`_ licensed modules for working with
points on Earth, or other near spherical objects.  It allows you to calculate
the distance and bearings between points, mangle xearth_/xplanet_ data files,
work with online UK trigpoint databases, NOAA_'s weather station database and
other such location databases.

.. _GPL v3: http://www.gnu.org/licenses/
.. _xearth: http://hewgill.com/xearth/original/
.. _xplanet: http://xplanet.sourceforge.net/
.. _NOAA: http://weather.noaa.gov/

The :mod:`upoints.point` module is the simplest interface available, and is
mainly useful as a naïve object for simple calculation and subclassing for
specific usage.  An example of how to use it follows:

>>> from upoints import point
>>> Home = point.Point(52.015, -0.221)
>>> Telford = point.Point(52.6333, -2.5000)
>>> int(Home.distance(Telford))
169
>>> int(Home.bearing(Telford))
294
>>> int(Home.final_bearing(Telford))
293
>>> import datetime
>>> Home.sun_events(datetime.date(2007, 6, 28))
(datetime.time(3, 42), datetime.time(20, 24))
>>> Home.sunrise(datetime.date(2007, 6, 28))
datetime.time(3, 42)
>>> Home.sunset(datetime.date(2007, 6, 28))
datetime.time(20, 24)

.. moduleauthor:: `%s <mailto:%s>`__
""" % parseaddr(__author__)

from upoints import (baken, cellid, cities, geonames, gpx, kml, nmea, osm,
                     point, trigpoints, tzdata, utils, weather_stations,
                     xearth)

__all__ = (baken, cellid, cities, geonames, gpx, kml, nmea, osm, point,
           trigpoints, tzdata, utils, weather_stations, xearth)



#
# coding=utf-8
"""cities - Imports GNU miscfiles cities data files"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__doc__ += """.

.. moduleauthor:: James Rowe <jnrowe@gmail.com>
.. versionadded:: 0.2.0
"""

import logging
import time

from upoints import (point, trigpoints, utils)

#: GNU miscfiles cities.dat template
TEMPLATE = """\
ID          : %s
Type        : %s
Population  : %s
Size        : %s
Name        : %s
 Country    : %s
 Region     : %s
Location    : %s
 Longitude  : %s
 Latitude   : %s
 Elevation  : %s
Date        : %s
Entered-By  : %s"""


class City(trigpoints.Trigpoint):

    """Class for representing an entry from the `GNU miscfiles`_ cities data file.

    .. versionadded:: 0.2.0

    .. _GNU miscfiles: http://directory.fsf.org/project/miscfiles/

    """

    __slots__ = ('identifier', 'ptype', 'population', 'size', 'country',
                 'region', 'location', 'date', 'entered')

    def __init__(self, identifier, name, ptype, region, country, location,
                 population, size, latitude, longitude, altitude, date,
                 entered):
        """Initialise a new ``City`` object.

        :param int identifier: Numeric identifier for object
        :param str name: Place name
        :param str ptype: Type of place
        :param str region: Region place can be found
        :param str country: Country name place can be found
        :param str location: Body place can be found
        :param int population: Place's population
        :param int size: Place's area
        :param float latitude: Station's latitude
        :param float longitude: Station's longitude
        :param int altitude: Station's elevation
        :param time.struct_time date: Date the entry was added
        :param str entered: Entry's author

        """
        super(City, self).__init__(latitude, longitude, altitude, name)
        self.identifier = identifier
        self.ptype = ptype
        self.region = region
        self.country = country
        self.location = location
        self.population = population
        self.size = size
        self.date = date
        self.entered = entered

    def __str__(self):
        """Pretty printed location string.

        :rtype: ``str``
        :return: Human readable string representation of ``City`` object

        """
        values = map(utils.value_or_empty,
                     (self.identifier, self.ptype,
                      self.population, self.size,
                      self.name, self.country,
                      self.region, self.location,
                      self.latitude, self.longitude,
                      self.altitude,
                      time.strftime('%Y%m%d', self.date) if self.date else '',
                      self.entered))
        return TEMPLATE % tuple(values)


class Cities(point.Points):

    """Class for representing a group of :class:`City` objects.

    .. versionadded:: 0.5.1

    """

    def __init__(self, data=None):
        """Initialise a new ``Cities`` object."""
        super(Cities, self).__init__()
        self._data = data
        if data:
            self.import_locations(data)

    def import_locations(self, data):
        """Parse `GNU miscfiles`_ cities data files.

        ``import_locations()`` returns a list containing :class:`City` objects.

        It expects data files in the same format that `GNU miscfiles`_
        provides, that is::

            ID          : 1
            Type        : City
            Population  : 210700
            Size        : 
            Name        : Aberdeen
             Country    : UK
             Region     : Scotland
            Location    : Earth
             Longitude  : -2.083
             Latitude   :   57.150
             Elevation  : 
            Date        : 19961206
            Entered-By  : Rob.Hooft@EMBL-Heidelberg.DE
            //
            ID          : 2
            Type        : City
            Population  : 1950000
            Size        : 
            Name        : Abidjan
             Country    : Ivory Coast
             Region     : 
            Location    : Earth
             Longitude  : -3.867
             Latitude   :    5.333
             Elevation  : 
            Date        : 19961206
            Entered-By  : Rob.Hooft@EMBL-Heidelberg.DE

        When processed by ``import_locations()`` will return ``list`` object in
        the following style::

            [City(1, "City", 210700, None, "Aberdeen", "UK", "Scotland",
                  "Earth", -2.083, 57.15, None, (1996, 12, 6, 0, 0, 0, 4,
                  341, -1), "Rob.Hooft@EMBL-Heidelberg.DE"),
             City(2, "City", 1950000, None, "Abidjan", "Ivory Coast", "",
                  "Earth", -3.867, 5.333, None, (1996, 12, 6, 0, 0, 0, 4,
                  341, -1), "Rob.Hooft@EMBL-Heidelberg.DE")])

        :type data: ``file``, ``list`` or ``str``
        :param data:
            :abbr:`NOAA (National Oceanographic and Atmospheric Administration)`
            station data to read
        :rtype: ``list``
        :return: Places as ``City`` objects
        :raise TypeError: Invalid value for data

        .. _GNU miscfiles: http://directory.fsf.org/project/miscfiles/

        """
        self._data = data
        if hasattr(data, 'read'):
            data = data.read().split('//\n')
        elif isinstance(data, list):
            pass
        elif isinstance(data, basestring):
            data = open(data).read().split('//\n')
        else:
            raise TypeError('Unable to handle data of type %r' % type(data))

        keys = ('identifier', 'ptype', 'population', 'size', 'name', 'country',
                'region', 'location', 'longitude', 'latitude', 'altitude',
                'date', 'entered')

        for record in data:
            # We truncate after splitting because the v1.4.2 datafile contains
            # a broken separator between 229 and 230 that would otherwise break
            # the import
            data = [i.split(':')[1].strip() for i in record.splitlines()[:13]]
            entries = dict(zip(keys, data))

            # Entry for Utrecht has the incorrect value of 0.000 for elevation.
            if entries['altitude'] == '0.000':
                logging.debug("Ignoring `0.000' value for elevation in %r "
                              'entry' % record)
                entries['altitude'] = ''
            for i in ('identifier', 'population', 'size', 'altitude'):
                entries[i] = int(entries[i]) if entries[i] else None
            for i in ('longitude', 'latitude'):
                entries[i] = float(entries[i]) if entries[i] else None
            entries['date'] = time.strptime(entries['date'], '%Y%m%d')
            self.append(City(**entries))



#
# coding=utf-8
"""test_cities - Test cities support"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import TestCase

from expecter import expect

from upoints.cities import (City, Cities)


class TestCity(TestCase):
    def setUp(self):
        self.t = City(498, 'Zwickau', 'City', 'Sachsen', 'DE', 'Earth', 108835,
                      None, 12.5, 50.72, None,
                      (1997, 4, 10, 0, 0, 0, 3, 100, -1),
                      'M.Dowling@tu-bs.de')

    def test___repr__(self):
        expect(repr(self.t)) == \
            ("City(498, 'Zwickau', 'City', 'Sachsen', 'DE', 'Earth', 108835, "
             'None, 12.5, 50.72, None, (1997, 4, 10, 0, 0, 0, 3, 100, -1), '
             "'M.Dowling@tu-bs.de')")

    def test___str__(self):
        data = str(self.t).splitlines()
        expect(data[0]) == 'ID          : 498'
        expect(data[1]) == 'Type        : City'
        expect(data[2]) == 'Population  : 108835'
        expect(data[3]) == 'Size        : '
        expect(data[4]) == 'Name        : Zwickau'
        expect(data[5]) == ' Country    : DE'
        expect(data[6]) == ' Region     : Sachsen'
        expect(data[7]) == 'Location    : Earth'
        expect(data[8]) == ' Longitude  : 12.5'
        expect(data[9]) == ' Latitude   : 50.72'
        expect(data[10]) == ' Elevation  : '
        expect(data[11]) == 'Date        : 19970410'
        expect(data[12]) == 'Entered-By  : M.Dowling@tu-bs.de'


class TestCities(TestCase):
    def test_import_locations(self):
        cities_file = open('tests/data/city_data')
        cities = Cities(cities_file)
        data = [('%i - %s (%s;%s)' % (city.identifier, city.name,
                                      city.latitude, city.longitude))
                for city in sorted(cities, key=lambda x: x.identifier)]
        expect(data[0]) == '126 - London (51.5;-0.083)'
        expect(data[1]) == '127 - Luxembourg (49.617;6.117)'
        expect(data[2]) == '128 - Lyon (45.767;4.867)'
        cities_file.seek(0)
        manual_list = cities_file.read().split('//\\n')
        cities = Cities(manual_list)
        expect(len(cities)) == 1



#
# coding=utf-8
"""utils - Utilities for test support"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from doctest import _ellipsis_match as ellipsis_match

from expecter import expect
from lxml import etree


def xml_compare(elem1, elem2, ellipsis=False):
    expect(elem1.tag) == elem2.tag
    for key, value in elem1.attrib.items():
        expect(elem2.attrib.get(key)) == value
    for key in elem2.attrib.keys():
        expect(elem1.attrib).contains(key)
    text1 = elem1.text.strip() if elem1.text else ''
    text2 = elem2.text.strip() if elem2.text else ''
    tail1 = elem1.tail.strip() if elem1.tail else ''
    tail2 = elem2.tail.strip() if elem2.tail else ''
    if ellipsis:
        if not ellipsis_match(text1, text2):
            raise ValueError('text mismatch: %r != %r' % (elem1.text,
                                                          elem2.text))
        if not ellipsis_match(tail1, tail2):
            raise ValueError('tail mismatch: %r != %r' % (elem1.text,
                                                          elem2.text))
    else:
        expect(text1) == text2
        expect(tail1) == tail2
    children1 = elem1.getchildren()
    children2 = elem2.getchildren()
    expect(len(children1)) == len(children2)
    for child1, child2 in zip(children1, children2):
        xml_compare(child1, child2, ellipsis)


def xml_str_compare(string1, string2, ellipsis=False):
    doc1 = etree.fromstring(string1)
    doc2 = etree.fromstring(string2)
    xml_compare(doc1, doc2, ellipsis)



#
# coding=utf-8
"""gpx - Imports GPS eXchange format data files"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import time

from operator import attrgetter

from lxml import etree

from upoints import (point, utils)
from upoints._version import web as ua_string


GPX_NS = 'http://www.topografix.com/GPX/1/1'
etree.register_namespace('gpx', GPX_NS)

create_elem = utils.element_creator(GPX_NS)

GPX_ELEM_ATTRIB = {
    'creator': ua_string,
    'version': '1.1',
    '{http://www.w3.org/2001/XMLSchema-instance}schemaLocation': '%s http://www.topografix.com/GPX/1/1/gpx.xsd' % GPX_NS,
}

class _GpxElem(point.TimedPoint):

    """Abstract class for representing an element from GPX data files.

    .. versionadded:: 0.11.0

    """

    __slots__ = ('name', 'description', 'elevation', 'time', )

    _elem_name = None

    def __init__(self, latitude, longitude, name=None, description=None,
                 elevation=None, time=None):
        """Initialise a new ``_GpxElem`` object.

        :param float latitude: Element's latitude
        :param float longitude: Element's longitude
        :param str name: Name for Element
        :param str description: Element's description
        :param float elevation: Element's elevation
        :param utils.Timestamp time: Time the data was generated

        """
        super(_GpxElem, self).__init__(latitude, longitude, time=time)
        self.name = name
        self.description = description
        self.elevation = elevation

    def __str__(self):
        """Pretty printed location string.

        :rtype: ``str``
        :return: Human readable string representation of :class:`_GpxElem`
            object

        """
        location = super(_GpxElem, self).__format__('dms')
        if self.elevation:
            location += ' @ %sm' % self.elevation
        if self.time:
            location += ' on %s' % self.time.isoformat()
        if self.name:
            text = ['%s (%s)' % (self.name, location), ]
        else:
            text = [location, ]
        if self.description:
            text.append('[%s]' % self.description)
        return ' '.join(text)

    def togpx(self):
        """Generate a GPX waypoint element subtree.

        :rtype: :class:`etree.Element`
        :return: GPX element

        """
        element = create_elem(self.__class__._elem_name,
                              {'lat': str(self.latitude),
                               'lon': str(self.longitude)})
        if self.name:
            element.append(create_elem('name', text=self.name))
        if self.description:
            element.append(create_elem('desc', text=self.description))
        if self.elevation:
            element.append(create_elem('ele', text=str(self.elevation)))
        if self.time:
            element.append(create_elem('time', text=self.time.isoformat()))
        return element


class _SegWrap(list):

    """Abstract class for representing segmented elements from GPX data files.

    .. versionadded:: 0.12.0

    """

    def __init__(self, gpx_file=None, metadata=None):
        """Initialise a new ``_SegWrap`` object."""
        super(_SegWrap, self).__init__()
        self.metadata = metadata if metadata else _GpxMeta()
        self._gpx_file = gpx_file
        if gpx_file:
            self.import_locations(gpx_file)

    def distance(self, method='haversine'):
        """Calculate distances between locations in segments.

        :Parameters:
            method : ``str``
                Method used to calculate distance

        :rtype: ``list`` of ``list`` of ``float``
        :return: Groups of distance between points in segments

        """
        distances = []
        for segment in self:
            if len(segment) < 2:
                distances.append([])
            else:
                distances.append(segment.distance(method))
        return distances

    def bearing(self, format='numeric'):
        """Calculate bearing between locations in segments.

        :param str format: Format of the bearing string to return
        :rtype: ``list`` of ``list`` of ``float``
        :return: Groups of bearings between points in segments

        """
        bearings = []
        for segment in self:
            if len(segment) < 2:
                bearings.append([])
            else:
                bearings.append(segment.bearing(format))
        return bearings

    def final_bearing(self, format='numeric'):
        """Calculate final bearing between locations in segments.

        :param str format: Format of the bearing string to return
        :rtype: ``list`` of ``list`` of ``float``
        :return: Groups of bearings between points in segments

        """
        bearings = []
        for segment in self:
            if len(segment) < 2:
                bearings.append([])
            else:
                bearings.append(segment.final_bearing(format))
        return bearings

    def inverse(self):
        """Calculate the inverse geodesic between locations in segments.

        :rtype: ``list`` of 2 ``tuple`` of ``float``
        :return: Groups in bearing and distance between points in segments

        """
        inverses = []
        for segment in self:
            if len(segment) < 2:
                inverses.append([])
            else:
                inverses.append(segment.inverse())
        return inverses

    def midpoint(self):
        """Calculate the midpoint between locations in segments.

        :rtype: ``list`` of ``Point`` instance
        :return: Groups of midpoint between points in segments

        """
        midpoints = []
        for segment in self:
            if len(segment) < 2:
                midpoints.append([])
            else:
                midpoints.append(segment.midpoint())
        return midpoints

    def range(self, location, distance):
        """Test whether locations are within a given range of ``location``.

        :param Point location: Location to test range against
        :param float distance: Distance to test location is within
        :rtype: ``list`` of ``list`` of ``Point`` objects within specified
            range
        :return: Groups of points in range per segment

        """
        return (segment.range(location, distance) for segment in self)

    def destination(self, bearing, distance):
        """Calculate destination locations for given distance and bearings.

        :param float bearing: Bearing to move on in degrees
        :param float distance: Distance in kilometres
        :rtype: ``list`` of ``list`` of ``Point``
        :return: Groups of points shifted by ``distance`` and ``bearing``

        """
        return (segment.destination(bearing, distance) for segment in self)
    forward = destination

    def sunrise(self, date=None, zenith=None):
        """Calculate sunrise times for locations.

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate sunrise events, or end of twilight
        :rtype: ``list`` of ``list`` of :class:`datetime.datetime`
        :return: The time for the sunrise for each point in each segment

        """
        return (segment.sunrise(date, zenith) for segment in self)

    def sunset(self, date=None, zenith=None):
        """Calculate sunset times for locations.

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate sunset events, or start of twilight times
        :rtype: ``list`` of ``list`` of :class:`datetime.datetime`
        :return: The time for the sunset for each point in each segment

        """
        return (segment.sunset(date, zenith) for segment in self)

    def sun_events(self, date=None, zenith=None):
        """Calculate sunrise/sunset times for locations.

        :param datetime.date date: Calculate rise or set for given date
        :param str zenith: Calculate rise/set events, or twilight times
        :rtype: ``list`` of ``list`` of 2 ``tuple`` of
            :class:`datetime.datetime`
        :return: The time for the sunrise and sunset events for each point in
            each segment

        """
        return (segment.sun_events(date, zenith) for segment in self)

    def to_grid_locator(self, precision='square'):
        """Calculate Maidenhead locator for locations.

        :param str precision: Precision with which generate locator string
        :rtype: ``list`` of ``list`` of ``str``
        :return: Groups of Maidenhead locator for each point in segments

        """
        return (segment.to_grid_locator(precision) for segment in self)

    def speed(self):
        """Calculate speed between locations per segment.

        :rtype: ``list`` of ``list`` of ``float``
        :return: Speed between points in each segment in km/h

        """
        return (segment.speed() for segment in self)


class _GpxMeta(object):

    """Class for representing GPX global metadata.

    .. versionadded:: 0.12.0

    """

    __slots__ = ('name', 'desc', 'author', 'copyright', 'link', 'time',
                 'keywords', 'bounds', 'extensions')

    def __init__(self, name=None, desc=None, author=None, copyright=None,
                 link=None, time=None, keywords=None, bounds=None,
                 extensions=None):
        """Initialise a new `_GpxMeta` object.

        :param str name: Name for the export
        :param str desc: Description for the GPX export
        :param dict author: Author of the entire GPX data
        :param dict copyright: Copyright data for the exported data
        :type link: ``list`` of ``str`` or ``dict``
        :param link: Links associated with the data
        :param utils.Timestamp time:Time the data was generated
        :param str keywords: Keywords associated with the data
        :type bounds: ``dict`` or ``list`` of ``Point`` objects
        :param bounds: Area used in the data
        :type extensions: ``list`` of :class:`etree.Element` objects
        :param extensions: Any external data associated with the export

        """
        super(_GpxMeta, self).__init__()
        self.name = name
        self.desc = desc
        self.author = author if author else {}
        self.copyright = copyright if copyright else {}
        self.link = link if link else []
        self.time = time
        self.keywords = keywords
        self.bounds = bounds
        self.extensions = extensions

    def togpx(self):
        """Generate a GPX metadata element subtree.

        :rtype: :class:`etree.Element`
        :return: GPX metadata element

        """
        metadata = create_elem('metadata')
        if self.name:
            metadata.append(create_elem('name', text=self.name))
        if self.desc:
            metadata.append(create_elem('desc', text=self.desc))
        if self.author:
            element = create_elem('author')
            if self.author['name']:
                element.append(create_elem('name', text=self.author['name']))
            if self.author['email']:
                attr = dict(zip(self.author['email'].split('@'),
                                ('id', 'domain')))
                element.append(create_elem('email', attr))
            if self.author['link']:
                element.append(create_elem('link', text=self.author['link']))
            metadata.append(element)
        if self.copyright:
            if self.copyright['name']:
                author = {'author': self.copyright['name']}
            else:
                author = None
            element = create_elem('copyright', author)
            if self.copyright['year']:
                element.append(create_elem('year', text=self.copyright['year']))
            if self.copyright['license']:
                license = create_elem('license')
                element.append(license)
            metadata.append(element)
        if self.link:
            for link in self.link:
                if isinstance(link, basestring):
                    element = create_elem('link', {'href': link})
                else:
                    element = create_elem('link', {'href': link['href']})
                    if link['text']:
                        element.append(create_elem('text', text=link['text']))
                    if link['type']:
                        element.append(create_elem('type', text=link['type']))
                metadata.append(element)
        if isinstance(self.time, (time.struct_time, tuple)):
            text = time.strftime('%Y-%m-%dT%H:%M:%S%z', self.time)
        elif isinstance(self.time, utils.Timestamp):
            text = self.time.isoformat()
        else:
            text = time.strftime('%Y-%m-%dT%H:%M:%S%z')
        metadata.append(create_elem('time', text=text))
        if self.keywords:
            metadata.append(create_elem('keywords', text=self.keywords))
        if self.bounds:
            if not isinstance(self.bounds, dict):
                latitudes = list(map(attrgetter('latitude'), self.bounds))
                longitudes = list(map(attrgetter('longitude'), self.bounds))
                bounds = {
                    'minlat': str(min(latitudes)),
                    'maxlat': str(max(latitudes)),
                    'minlon': str(min(longitudes)),
                    'maxlon': str(max(longitudes)),
                }
            else:
                bounds = dict((k, str(v)) for k, v in self.bounds.items())
            metadata.append(create_elem('bounds', bounds))
        if self.extensions:
            element = create_elem('extensions')
            for i in self.extensions:
                element.append(i)
            metadata.append(self.extensions)
        return metadata

    def import_metadata(self, elements):
        """Import information from GPX metadata.

        :param etree.Element elements: GPX metadata subtree

        """
        metadata_elem = lambda name: etree.QName(GPX_NS, name)

        for child in elements.getchildren():
            tag_ns, tag_name = child.tag[1:].split('}')
            if not tag_ns == GPX_NS:
                continue
            if tag_name in ('name', 'desc', 'keywords'):
                setattr(self, tag_name, child.text)
            elif tag_name == 'time':
                self.time = utils.Timestamp.parse_isoformat(child.text)
            elif tag_name == 'author':
                self.author['name'] = child.findtext(metadata_elem('name'))
                aemail = child.find(metadata_elem('email'))
                if aemail:
                    self.author['email'] = '%s@%s' % (aemail.get('id'),
                                                      aemail.get('domain'))
                self.author['link'] = child.findtext(metadata_elem('link'))
            elif tag_name == 'bounds':
                self.bounds = {
                    'minlat': child.get('minlat'),
                    'maxlat': child.get('maxlat'),
                    'minlon': child.get('minlon'),
                    'maxlon': child.get('maxlon'),
                }
            elif tag_name == 'extensions':
                self.extensions = child.getchildren()
            elif tag_name == 'copyright':
                if child.get('author'):
                    self.copyright['name'] = child.get('author')
                self.copyright['year'] = child.findtext(metadata_elem('year'))
                self.copyright['license'] = child.findtext(metadata_elem('license'))
            elif tag_name == 'link':
                link = {
                    'href': child.get('href'),
                    'type': child.findtext(metadata_elem('type')),
                    'text': child.findtext(metadata_elem('text')),
                }
                self.link.append(link)


class Waypoint(_GpxElem):

    """Class for representing a waypoint element from GPX data files.

    .. versionadded:: 0.8.0

    .. seealso::

        :class:`_GpxElem`

    """

    __slots__ = ('name', 'description', )

    _elem_name = 'wpt'


class Waypoints(point.TimedPoints):

    """Class for representing a group of :class:`Waypoint` objects.

    .. versionadded:: 0.8.0

    """

    def __init__(self, gpx_file=None, metadata=None):
        """Initialise a new ``Waypoints`` object."""
        super(Waypoints, self).__init__()
        self.metadata = metadata if metadata else _GpxMeta()
        self._gpx_file = gpx_file
        if gpx_file:
            self.import_locations(gpx_file)

    def import_locations(self, gpx_file):
        """Import GPX data files.

        ``import_locations()`` returns a list with :class:`Waypoint` objects.

        It expects data files in GPX format, as specified in `GPX 1.1 Schema
        Documentation`_, which is XML such as::

            <?xml version="1.0" encoding="utf-8" standalone="no"?>
            <gpx version="1.1" creator="PocketGPSWorld.com"
            xmlns="http://www.topografix.com/GPX/1/1"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xsi:schemaLocation="http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd">

              <wpt lat="52.015" lon="-0.221">
                <name>Home</name>
                <desc>My place</desc>
              </wpt>
              <wpt lat="52.167" lon="0.390">
                <name>MSR</name>
                <desc>Microsoft Research, Cambridge</desc>
              </wpt>
            </gpx>

        The reader uses the :mod:`ElementTree` module, so should be very fast
        when importing data.  The above file processed by ``import_locations()``
        will return the following ``list`` object::

            [Waypoint(52.015, -0.221, "Home", "My place"),
             Waypoint(52.167, 0.390, "MSR", "Microsoft Research, Cambridge")]

        :type gpx_file: ``file``, ``list`` or ``str``
        :param gpx_file: GPX data to read
        :rtype: ``list``
        :return: Locations with optional comments

        .. _GPX 1.1 Schema Documentation: http://www.topografix.com/GPX/1/1/

        """
        self._gpx_file = gpx_file
        data = utils.prepare_xml_read(gpx_file, objectify=True)

        try:
            self.metadata.import_metadata(data.metadata)
        except AttributeError:
            pass

        for waypoint in data.wpt:
            latitude = waypoint.get('lat')
            longitude = waypoint.get('lon')
            try:
                name = waypoint.name.text
            except AttributeError:
                name = None
            try:
                description = waypoint.desc.text
            except AttributeError:
                description = None
            try:
                elevation = float(waypoint.ele.text)
            except AttributeError:
                elevation = None
            try:
                time = utils.Timestamp.parse_isoformat(waypoint.time.text)
            except AttributeError:
                time = None
            self.append(Waypoint(latitude, longitude, name, description,
                                 elevation, time))

    def export_gpx_file(self):
        """Generate GPX element tree from ``Waypoints`` object.

        :rtype: :class:`etree.ElementTree`
        :return: GPX element tree depicting ``Waypoints`` object

        """
        gpx = create_elem('gpx', GPX_ELEM_ATTRIB)
        if not self.metadata.bounds:
            self.metadata.bounds = self[:]
        gpx.append(self.metadata.togpx())
        for place in self:
            gpx.append(place.togpx())

        return etree.ElementTree(gpx)


class Trackpoint(_GpxElem):

    """Class for representing a trackpoint element from GPX data files.

    .. versionadded:: 0.10.0

    .. seealso::

        :class:`_GpxElem`

    """

    __slots__ = ('name', 'description', )

    _elem_name = 'trkpt'


class Trackpoints(_SegWrap):

    """Class for representing a group of :class:`Trackpoint` objects.

    .. versionadded:: 0.10.0

    """

    def import_locations(self, gpx_file):
        """Import GPX data files.

        ``import_locations()`` returns a series of lists representing track
        segments with :class:`Trackpoint` objects as contents.

        It expects data files in GPX format, as specified in `GPX 1.1 Schema
        Documentation`_, which is XML such as::

            <?xml version="1.0" encoding="utf-8" standalone="no"?>
            <gpx version="1.1" creator="upoints/0.12.0"
            xmlns="http://www.topografix.com/GPX/1/1"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xsi:schemaLocation="http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd">
              <trk>
                <trkseg>
                  <trkpt lat="52.015" lon="-0.221">
                    <name>Home</name>
                    <desc>My place</desc>
                  </trkpt>
                  <trkpt lat="52.167" lon="0.390">
                    <name>MSR</name>
                    <desc>Microsoft Research, Cambridge</desc>
                  </trkpt>
                </trkseg>
              </trk>
            </gpx>

        The reader uses the :mod:`ElementTree` module, so should be very fast
        when importing data.  The above file processed by ``import_locations()``
        will return the following ``list`` object::

            [[Trackpoint(52.015, -0.221, "Home", "My place"),
              Trackpoint(52.167, 0.390, "MSR", "Microsoft Research, Cambridge")], ]

        :type gpx_file: ``file``, ``list`` or ``str``
        :param gpx_file: GPX data to read
        :rtype: ``list``
        :return: Locations with optional comments

        .. _GPX 1.1 Schema Documentation: http://www.topografix.com/GPX/1/1/

        """
        self._gpx_file = gpx_file
        data = utils.prepare_xml_read(gpx_file, objectify=True)

        try:
            self.metadata.import_metadata(data.metadata)
        except AttributeError:
            pass

        for segment in data.trk.trkseg:
            points = point.TimedPoints()
            for trackpoint in segment.trkpt:
                latitude = trackpoint.get('lat')
                longitude = trackpoint.get('lon')
                try:
                    name = trackpoint.name.text
                except AttributeError:
                    name = None
                try:
                    description = trackpoint.desc.text
                except AttributeError:
                    description = None
                try:
                    elevation = float(trackpoint.ele.text)
                except AttributeError:
                    elevation = None
                try:
                    time = utils.Timestamp.parse_isoformat(trackpoint.time.text)
                except AttributeError:
                    time = None
                points.append(Trackpoint(latitude, longitude, name,
                                         description, elevation, time))
            self.append(points)

    def export_gpx_file(self):
        """Generate GPX element tree from ``Trackpoints``.

        :rtype: :class:`etree.ElementTree`
        :return: GPX element tree depicting ``Trackpoints`` objects

        """
        gpx = create_elem('gpx', GPX_ELEM_ATTRIB)
        if not self.metadata.bounds:
            self.metadata.bounds = [j for i in self for j in i]
        gpx.append(self.metadata.togpx())
        track = create_elem('trk')
        gpx.append(track)
        for segment in self:
            chunk = create_elem('trkseg')
            track.append(chunk)
            for place in segment:
                chunk.append(place.togpx())

        return etree.ElementTree(gpx)


class Routepoint(_GpxElem):

    """Class for representing a ``rtepoint`` element from GPX data files.

    .. versionadded:: 0.10.0

    .. seealso:

         :class:`_GpxElem`

    """

    __slots__ = ('name', 'description', )

    _elem_name = 'rtept'


class Routepoints(_SegWrap):

    """Class for representing a group of :class:`Routepoint` objects.

    .. versionadded:: 0.10.0

    """

    def import_locations(self, gpx_file):
        """Import GPX data files.

        ``import_locations()`` returns a series of lists representing track
        segments with :class:`Routepoint` objects as contents.

        It expects data files in GPX format, as specified in `GPX 1.1 Schema
        Documentation`_, which is XML such as::

            <?xml version="1.0" encoding="utf-8" standalone="no"?>
            <gpx version="1.1" creator="upoints/0.12.0"
            xmlns="http://www.topografix.com/GPX/1/1"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xsi:schemaLocation="http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd">
              <rte>
                <rtept lat="52.015" lon="-0.221">
                  <name>Home</name>
                  <desc>My place</desc>
                </rtept>
                <rtept lat="52.167" lon="0.390">
                  <name>MSR</name>
                  <desc>Microsoft Research, Cambridge</desc>
                </rtept>
              </rte>
            </gpx>

        The reader uses the :mod:`ElementTree` module, so should be very fast
        when importing data.  The above file processed by ``import_locations()``
        will return the following ``list`` object::

            [[Routepoint(52.015, -0.221, "Home", "My place"),
              Routepoint(52.167, 0.390, "MSR", "Microsoft Research, Cambridge")], ]

        :type gpx_file: ``file``, ``list`` or ``str``
        :param gpx_file: GPX data to read
        :rtype: ``list``
        :return: Locations with optional comments

        .. _GPX 1.1 Schema Documentation: http://www.topografix.com/GPX/1/1/

        """
        self._gpx_file = gpx_file
        data = utils.prepare_xml_read(gpx_file, objectify=True)

        try:
            self.metadata.import_metadata(data.metadata)
        except AttributeError:
            pass

        for route in data.rte:
            points = point.TimedPoints()
            for routepoint in route.rtept:
                latitude = routepoint.get('lat')
                longitude = routepoint.get('lon')
                try:
                    name = routepoint.name.text
                except AttributeError:
                    name = None
                try:
                    description = routepoint.desc.text
                except AttributeError:
                    description = None
                try:
                    elevation = float(routepoint.ele.text)
                except AttributeError:
                    elevation = None
                try:
                    time = utils.Timestamp.parse_isoformat(routepoint.time.text)
                except AttributeError:
                    time = None
                points.append(Routepoint(latitude, longitude, name,
                                         description, elevation, time))
            self.append(points)

    def export_gpx_file(self):
        """Generate GPX element tree from :class:`Routepoints`

        :rtype: :class:`etree.ElementTree`
        :return: GPX element tree depicting :class:`Routepoints` objects

        """
        gpx = create_elem('gpx', GPX_ELEM_ATTRIB)
        if not self.metadata.bounds:
            self.metadata.bounds = [j for i in self for j in i]
        gpx.append(self.metadata.togpx())
        for rte in self:
            chunk = create_elem('rte')
            gpx.append(chunk)
            for place in rte:
                chunk.append(place.togpx())

        return etree.ElementTree(gpx)



#
# coding=utf-8
"""weather_stations - Imports weather station data files"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__doc__ += """.

.. moduleauthor:: James Rowe <jnrowe@gmail.com>
.. versionadded:: 0.2.0
"""


import logging

from upoints import (point, trigpoints, utils)


class Station(trigpoints.Trigpoint):

    """Class for representing a weather station from a NOAA data file.

    .. versionadded:: 0.2.0

    """

    __slots__ = ('alt_id', 'state', 'country', 'wmo', 'ua_latitude',
                 'ua_longitude', 'ua_altitude', 'rbsn')

    def __init__(self, alt_id, name, state, country, wmo, latitude, longitude,
                 ua_latitude, ua_longitude, altitude, ua_altitude, rbsn):
        """Initialise a new ``Station`` object.

        :param str alt_id: Alternate location identifier
        :param str name: Station's name
        :param str state: State name, if station is in the US
        :param str country: Country name
        :param int wmo: WMO region code
        :param float latitude: Station's latitude
        :param float longitude: Station's longitude
        :param float ua_latitude: Station's upper air latitude
        :param float ua_longitude: Station's upper air longitude
        :param int altitude: Station's elevation
        :param int ua_altitude: Station's upper air elevation
        :param bool rbsn: True if station belongs to RSBN

        """
        super(Station, self).__init__(latitude, longitude, altitude, name)
        self.alt_id = alt_id
        self.state = state
        self.country = country
        self.wmo = wmo
        self.ua_latitude = ua_latitude
        self.ua_longitude = ua_longitude
        self.ua_altitude = ua_altitude
        self.rbsn = rbsn

    def __str__(self):
        """Pretty printed location string.

        .. seealso::

           :type :class:`trigpoints.point.Point`

        :rtype: ``str``
        :return: Human readable string representation of ``Station`` object

        """
        return self.__format__()

    def __format__(self, format_spec='dd'):
        """Extended pretty printing for location strings.

        :param str format_spec: Coordinate formatting system to use
        :rtype: ``str``
        :return: Human readable string representation of ``Point`` object
        :raise ValueError: Unknown value for ``format_spec``

        """
        text = super(Station.__base__, self).__format__(format_spec)

        if self.alt_id:
            return '%s (%s - %s)' % (self.name, self.alt_id, text)
        else:
            return '%s (%s)' % (self.name, text)


class Stations(point.KeyedPoints):

    """Class for representing a group of `Station` objects.

    .. versionadded:: 0.5.1

    """

    def __init__(self, data=None, index='WMO'):
        """Initialise a new `Stations` object."""
        super(Stations, self).__init__()
        self._data = data
        self._index = index
        if data:
            self.import_locations(data, index)

    def import_locations(self, data, index='WMO'):
        """Parse NOAA weather station data files.

        ``import_locations()`` returns a dictionary with keys containing either
        the WMO or ICAO identifier, and values that are ``Station`` objects
        that describes the large variety of data exported by NOAA_.

        It expects data files in one of the following formats::

            00;000;PABL;Buckland, Buckland Airport;AK;United States;4;65-58-56N;161-09-07W;;;7;;
            01;001;ENJA;Jan Mayen;;Norway;6;70-56N;008-40W;70-56N;008-40W;10;9;P
            01;002;----;Grahuken;;Norway;6;79-47N;014-28E;;;;15;

        or::

            AYMD;94;014;Madang;;Papua New Guinea;5;05-13S;145-47E;05-13S;145-47E;3;5;P
            AYMO;--;---;Manus Island/Momote;;Papua New Guinea;5;02-03-43S;147-25-27E;;;4;;
            AYPY;94;035;Moresby;;Papua New Guinea;5;09-26S;147-13E;09-26S;147-13E;38;49;P

        Files containing the data in this format can be downloaded from the
        :abbr:`NOAA (National Oceanographic and Atmospheric Administration)`'s
        site in their `station location page`_.

        WMO indexed files downloaded from the :abbr:`NOAA (National
        Oceanographic and Atmospheric Administration)` site when processed by
        ``import_locations()`` will return ``dict`` object of the following
        style::

            {'00000': Station('PABL', 'Buckland, Buckland Airport', 'AK',
                              'United States', 4, 65.982222. -160.848055, None,
                              None, 7, False),
             '01001'; Station('ENJA', Jan Mayen, None, 'Norway', 6, 70.933333,
                              -7.333333, 70.933333, -7.333333, 10, 9, True),
             '01002': Station(None, 'Grahuken', None, 'Norway', 6, 79.783333,
                              13.533333, None, None, 15, False)}

        And ``dict`` objects such as the following will be created when ICAO
        indexed data files are processed::

            {'AYMD': Station("94", "014", "Madang", None, "Papua New Guinea",
                             5, -5.216666, 145.783333, -5.216666,
                             145.78333333333333, 3, 5, True,
             'AYMO': Station(None, None, "Manus Island/Momote", None,
                             "Papua New Guinea", 5, -2.061944, 147.424166,
                             None, None, 4, False,
             'AYPY': Station("94", "035", "Moresby", None, "Papua New Guinea",
                             5, -9.433333, 147.216667, -9.433333, 147.216667,
                             38, 49, True}

        :type data: ``file``, ``list`` or ``str``
        :param data: NOAA station data to read
        :param str index: The identifier type used in the file
        :rtype: ``dict``
        :return: WMO locations with `Station` objects
        :raise FileFormatError: Unknown file format

        .. _NOAA: http://weather.noaa.gov/
        .. _station location page: http://weather.noaa.gov/tg/site.shtml

        """
        self._data = data
        data = utils.prepare_read(data)

        for line in data:
            line = line.strip()
            chunk = line.split(';')
            if not len(chunk) == 14:
                if index == 'ICAO':
                    # Some entries only have 12 or 13 elements, so we assume 13
                    # and 14 are None.  Of the entries I've hand checked this
                    # assumption would be correct.
                    logging.debug('Extending ICAO %r entry, because it is '
                                  'too short to process' % line)
                    chunk.extend(['', ''])
                elif index == 'WMO' and len(chunk) == 13:
                    # A few of the WMO indexed entries are missing their RBSN
                    # fields, hand checking the entries for 71046 and 71899
                    # shows that they are correct if we just assume RBSN is
                    # false.
                    logging.debug('Extending WMO %r entry, because it is '
                                  'too short to process' % line)
                    chunk.append('')
                else:
                    raise utils.FileFormatError('NOAA')
            if index == 'WMO':
                identifier = ''.join(chunk[:2])
                alt_id = chunk[2]
            elif index == 'ICAO':
                identifier = chunk[0]
                alt_id = ''.join(chunk[1:3])
            else:
                raise ValueError('Unknown format %r' % index)
            if alt_id in ('----', '-----'):
                alt_id = None
            name = chunk[3]
            state = chunk[4] if chunk[4] else None
            country = chunk[5]
            wmo = int(chunk[6]) if chunk[6] else None
            point_data = []
            for i in chunk[7:11]:
                if not i:
                    point_data.append(None)
                    continue
                # Some entries in nsd_cccc.txt are of the format "DD-MM-
                # N", so we just take the spaces to mean 0 seconds.
                if ' ' in i:
                    logging.debug('Fixing unpadded location data in %r entry'
                                  % line)
                    i = i.replace(' ', '0')
                values = map(int, i[:-1].split('-'))
                if i[-1] in ('S', 'W'):
                    values = [-i for i in values]
                point_data.append(point.utils.to_dd(*values))
            latitude, longitude, ua_latitude, ua_longitude = point_data
            altitude = int(chunk[11]) if chunk[11] else None
            ua_altitude = int(chunk[12]) if chunk[12] else None
            rbsn = False if not chunk[13] else True
            self[identifier] = Station(alt_id, name, state, country, wmo,
                                       latitude, longitude, ua_latitude,
                                       ua_longitude, altitude, ua_altitude,
                                       rbsn)



#
# coding=utf-8
"""trigpoints - Imports trigpoint marker files"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__doc__ += """.

.. moduleauthor:: James Rowe <jnrowe@gmail.com>
.. versionadded:: 0.2.0
"""

from functools import partial

from upoints import (point, utils)


class Trigpoint(point.Point):

    """Class for representing a location from a trigpoint marker file.

    .. warning::
       Although this class stores and presents the representation of altitude
       it doesn't take it in to account when making calculations.  For example,
       consider a point at the base of Mount Everest and a point at the peak of
       Mount Everest the actual distance travelled between the two would be
       considerably larger than the reported value calculated at ground level.

    .. versionadded:: 0.2.0

    """

    __slots__ = ('altitude', 'name', 'identity')

    def __init__(self, latitude, longitude, altitude, name=None,
                 identity=None):
        """Initialise a new ``Trigpoint`` object.

        :param float latitude: Location's latitude
        :param float longitude: Location's longitude
        :param float altitude: Location's altitude
        :param str name: Name for location
        :param int identity: Database identifier, if known

        """
        super(Trigpoint, self).__init__(latitude, longitude)
        self.altitude = altitude
        self.name = name
        self.identity = identity

    def __str__(self):
        """Pretty printed location string.

        .. seealso::

           :type :class:`trigpoints.point.Point`

        :rtype: ``str``
        :return: Human readable string representation of ``Station`` object

        """
        return self.__format__()

    def __format__(self, format_spec='dms'):
        """Extended pretty printing for location strings.

        :param str format_spec: Coordinate formatting system to use
        :rtype: ``str``
        :return: Human readable string representation of ``Trigpoint`` object
        :raise ValueError: Unknown value for ``format_spec``

        """
        location = [super(Trigpoint, self).__format__(format_spec), ]
        if self.altitude:
            location.append('alt %im' % self.altitude)

        if self.name:
            return '%s (%s)' % (self.name, ' '.join(location))
        else:
            return ' '.join(location)


class Trigpoints(point.KeyedPoints):

    """Class for representing a group of :class:`Trigpoint` objects.

    .. versionadded:: 0.5.1

    """

    def __init__(self, marker_file=None):
        """Initialise a new ``Trigpoints`` object."""
        super(Trigpoints, self).__init__()
        self._marker_file = marker_file
        if marker_file:
            self.import_locations(marker_file)

    def import_locations(self, marker_file):
        """Import trigpoint database files.

        ``import_locations()`` returns a dictionary with keys containing the
        trigpoint identifier, and values that are :class:`Trigpoint` objects.

        It expects trigpoint marker files in the format provided at
        alltrigs-wgs84.txt_, which is the following format::

            H  SOFTWARE NAME & VERSION
            I  GPSU 4.04,
            S SymbolSet=0
            ...
            W,500936,N52.066035,W000.281449,    37.0,Broom Farm
            W,501097,N52.010585,W000.173443,    97.0,Bygrave
            W,505392,N51.910886,W000.186462,   136.0,Sish Lane

        Any line not consisting of 6 comma separated fields will be ignored.
        The reader uses the :mod:`csv` module, so alternative whitespace
        formatting should have no effect.  The above file processed by
        ``import_locations()`` will return the following ``dict`` object::

            {500936: point.Point(52.066035, -0.281449, 37.0, "Broom Farm"),
             501097: point.Point(52.010585, -0.173443, 97.0, "Bygrave"),
             505392: point.Point(51.910886, -0.186462, 136.0, "Sish Lane")}

        :type marker_file: ``file``, ``list`` or ``str``
        :param marker_file: Trigpoint marker data to read
        :rtype: ``dict``
        :return: Named locations with :class:`Trigpoint` objects
        :raise ValueError: Invalid value for ``marker_file``

        .. _alltrigs-wgs84.txt: http://www.haroldstreet.org.uk/trigpoints/

        """
        self._marker_file = marker_file
        field_names = ('tag', 'identity', 'latitude', 'longitude', 'altitude',
                       'name')
        pos_parse = lambda x, s: float(s[1:]) if s[0] == x else 0 - float(s[1:])
        latitude_parse = partial(pos_parse, 'N')
        longitude_parse = partial(pos_parse, 'E')
        # A value of 8888.0 denotes unavailable data
        altitude_parse = lambda s: None if s.strip() == '8888.0' else float(s)
        field_parsers = (str, int, latitude_parse, longitude_parse,
                         altitude_parse, str)

        data = utils.prepare_csv_read(marker_file, field_names)

        for row in (x for x in data if x['tag'] == 'W'):
            for name, parser in zip(field_names, field_parsers):
                row[name] = parser(row[name])
            del row['tag']
            try:
                self[row['identity']] = Trigpoint(**row)
            except TypeError:
                # Workaround formatting error in 506514 entry that contains
                # spurious comma
                del row[None]
                self[row['identity']] = Trigpoint(**row)



#
# coding=utf-8
"""utils - Support code for :mod:upoints"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division

#: Address for use in messages
__bug_report__ = 'James Rowe <jnrowe@gmail.com>'

import csv
import datetime
import inspect
import math
import re

from functools import reduce

from lxml import etree
from lxml import objectify as _objectify

from operator import add

from upoints.compat import (basestring, mangle_repr_type)


#: Body radii of various solar system objects
BODIES = {
    # Body radii in kilometres
    'Sun': 696000,

    # Terrestrial inner planets
    'Mercury': 2440,
    'Venus': 6052,
    'Earth': 6367,
    'Mars': 3390,

    # Gas giant outer planets
    'Jupiter': 69911,
    'Saturn': 58232,
    'Uranus': 25362,
    'Neptune': 24622,

    # only satellite to be added
    'Moon': 1738,

    # dwarf planets may be added
    'Pluto': 1153,
    'Ceres': 475,
    'Eris': 1200,
}

#: Default body radius to use for calculations
BODY_RADIUS = BODIES['Earth']
#: Number of kilometres per nautical mile
NAUTICAL_MILE = 1.852
#: Number of kilometres per statute mile
STATUTE_MILE = 1.609

#: Maidenhead locator constants
LONGITUDE_FIELD = 20
LATITUDE_FIELD = 10
LONGITUDE_SQUARE = LONGITUDE_FIELD / 10
LATITUDE_SQUARE = LATITUDE_FIELD / 10
LONGITUDE_SUBSQUARE = LONGITUDE_SQUARE / 24
LATITUDE_SUBSQUARE = LATITUDE_SQUARE / 24
LONGITUDE_EXTSQUARE = LONGITUDE_SUBSQUARE / 10
LATITUDE_EXTSQUARE = LATITUDE_SUBSQUARE / 10


class FileFormatError(ValueError):

    """Error object for data parsing error.

    .. versionadded:: 0.3.0

    """

    def __init__(self, site=None):
        """Initialise a new ``FileFormatError`` object.

        :param str site: Remote site name to display in error message

        """
        super(FileFormatError, self).__init__()
        self.site = site

    def __str__(self):
        """Pretty printed error string.

        :rtype: ``str``
        :return: Human readable error string

        """
        if self.site:
            return ("Incorrect data format, if you're using a file downloaded "
                    'from %s please report this to %s' % (self.site,
                                                          __bug_report__))
        else:
            return 'Unsupported data format.'


#{ Implementation utilities
def value_or_empty(value):
    """Return an empty string for display when value is ``None``.

    :param str value: Value to prepare for display
    :rtype: ``str``
    :return: String representation of ``value``

    """
    return value if value else ''


def repr_assist(obj, remap=None):
    """Helper function to simplify ``__repr__`` methods.

    :param obj: Object to pull argument values for
    :param dict remap: Argument pairs to remap before output
    :rtype: ``str``
    :return: Self-documenting representation of ``value``

    """
    if not remap:
        remap = {}
    data = []
    for arg in inspect.getargspec(getattr(obj.__class__, '__init__'))[0]:
        if arg == 'self':
            continue
        elif arg in remap:
            value = remap[arg]
        else:
            try:
                value = getattr(obj, arg)
            except AttributeError:
                value = getattr(obj, '_%s' % arg)
        if isinstance(value, (type(None), list, basestring, datetime.date,
                              datetime.time)):
            data.append(repr(value))
        else:
            data.append(str(value))
    return u"%s(%s)" % (obj.__class__.__name__, u', '.join(data))


def prepare_read(data, method='readlines', mode='r'):
    """Prepare various input types for parsing.

    :type data: ``file`` like object, ``list``, ``str``
    :param data: Data to read
    :param str method: Method to process data with
    :param str mode: Custom mode to process with, if data is a file
    :rtype: ``list``
    :return: List suitable for parsing
    :raise TypeError: Invalid value for data

    """
    if hasattr(data, 'readlines'):
        data = getattr(data, method)()
    elif isinstance(data, list):
        if method == 'read':
            return ''.join(data)
    elif isinstance(data, basestring):
        data = getattr(open(data, mode), method)()
    else:
        raise TypeError('Unable to handle data of type %r' % type(data))
    return data


def prepare_csv_read(data, field_names, *args, **kwargs):
    """Prepare various input types for CSV parsing.

    :type data: ``file`` like object, ``list``, ``str``
    :param data: Data to read
    :type field_names: ``tuple`` of ``str``
    :param field_names: Ordered names to assign to fields
    :rtype: `csv.DictReader`
    :return: CSV reader suitable for parsing
    :raise TypeError: Invalid value for data

    """
    if hasattr(data, 'readlines') or isinstance(data, list):
        pass
    elif isinstance(data, basestring):
        data = open(data)
    else:
        raise TypeError('Unable to handle data of type %r' % type(data))
    return csv.DictReader(data, field_names, *args, **kwargs)


def prepare_xml_read(data, objectify=False):
    """Prepare various input types for XML parsing.

    :type data: ``file`` like object, ``list``, ``str``
    :param data: Data to read
    :type objectify: bool
    :param objectify: Parse using lxml's objectify data binding
    :rtype: :class:`etree.ElementTree`
    :return: Tree suitable for parsing
    :raise TypeError: Invalid value for data

    """
    mod = _objectify if objectify else etree
    if hasattr(data, 'readlines'):
        data = mod.parse(data).getroot()
    elif isinstance(data, list):
        data = mod.fromstring(''.join(data))
    elif isinstance(data, basestring):
        data = mod.parse(open(data)).getroot()
    else:
        raise TypeError('Unable to handle data of type %r' % type(data))
    return data


def element_creator(namespace=None):
    """Create a simple namespace-aware objectify element creator.

    :param str namespace: Namespace to work in
    :rtype: ``function``
    :return: Namespace-aware element creator

    """
    ELEMENT_MAKER = _objectify.ElementMaker(namespace=namespace,
                                            annotate=False)

    def create_elem(tag, attr=None, text=None):
        """:class:`objectify.Element` wrapper with namespace defined.

        :param str tag: Tag name
        :param dict attr: Default attributes for tag
        :param str text: Text content for the tag
        :rtype: ``_objectify.ObjectifiedElement``
        :return: objectify element

        """
        if not attr:
            attr = {}
        if text:
            element = getattr(ELEMENT_MAKER, tag)(text, **attr)
        else:
            element = getattr(ELEMENT_MAKER, tag)(**attr)
        return element

    return create_elem

#}


#{ Angle conversion utilities
def to_dms(angle, style='dms'):
    """Convert decimal angle to degrees, minutes and possibly seconds.

    :param float angle: Angle to convert
    :param str style: Return fractional or whole minutes values
    :rtype: ``tuple`` of ``int`` objects for values
    :return: Angle converted to degrees, minutes and possibly seconds
    :raise ValueError: Unknown value for ``style``

    """
    sign = 1 if angle >= 0 else -1
    angle = abs(angle) * 3600
    minutes, seconds = divmod(angle, 60)
    degrees, minutes = divmod(minutes, 60)
    if style == 'dms':
        return tuple(sign * abs(i) for i in (int(degrees), int(minutes),
                                             seconds))
    elif style == 'dm':
        return tuple(sign * abs(i) for i in (int(degrees),
                                             (minutes + seconds / 60)))
    else:
        raise ValueError('Unknown style type %r' % style)


def to_dd(degrees, minutes, seconds=0):
    """Convert degrees, minutes and optionally seconds to decimal angle.

    :param float degrees: Number of degrees
    :param float minutes: Number of minutes
    :param float seconds: Number of seconds
    :rtype: ``float``
    :return: Angle converted to decimal degrees

    """
    sign = -1 if any(i < 0 for i in (degrees, minutes, seconds)) else 1
    return sign * (abs(degrees) + abs(minutes) / 60 + abs(seconds) / 3600)


def __chunk(segment, abbr=False):
    """Generate a ``tuple`` of compass direction names.

    :param list segment: Compass segment to generate names for
    :param bool abbr: Names should use single letter abbreviations
    :rtype: ``tuple``
    :return: Direction names for compass segment

    """
    names = ('north', 'east', 'south', 'west', 'north')
    if not abbr:
        sjoin = '-'
    else:
        names = [s[0].upper() for s in names]
        sjoin = ''
    if segment % 2 == 0:
        return (names[segment].capitalize(),
                sjoin.join((names[segment].capitalize(), names[segment],
                            names[segment + 1])),
                sjoin.join((names[segment].capitalize(), names[segment + 1])),
                sjoin.join((names[segment + 1].capitalize(), names[segment],
                            names[segment + 1])))
    else:
        return (names[segment].capitalize(),
                sjoin.join((names[segment].capitalize(), names[segment + 1],
                            names[segment])),
                sjoin.join((names[segment + 1].capitalize(), names[segment])),
                sjoin.join((names[segment + 1].capitalize(),
                            names[segment + 1], names[segment])))
COMPASS_NAMES = reduce(add, map(__chunk, range(4)))
COMPASS_NAMES_ABBR = reduce(add, [__chunk(x, True) for x in range(4)])


def angle_to_name(angle, segments=8, abbr=False):
    """Convert angle in to direction name.

    :param float angle: Angle in degrees to convert to direction name
    :param int segments: Number of segments to split compass in to
    :param bool abbr: Whether to return abbreviated direction string
    :rtype: ``str``
    :return: Direction name for ``angle``

    """
    if segments == 4:
        string = COMPASS_NAMES[int((angle + 45) / 90) % 4 * 2]
    elif segments == 8:
        string = COMPASS_NAMES[int((angle + 22.5) / 45) % 8 * 2]
    elif segments == 16:
        string = COMPASS_NAMES[int((angle + 11.25) / 22.5) % 16]
    else:
        raise ValueError('Segments parameter must be 4, 8 or 16 not %r'
                         % segments)
    if abbr:
        return ''.join(i[0].capitalize() for i in string.split('-'))
    else:
        return string
#}


#{ Date and time handling utilities
@mangle_repr_type
class TzOffset(datetime.tzinfo):

    """Time offset from UTC."""

    def __init__(self, tzstring):
        """Initialise a new ``TzOffset`` object.

        :param str tzstring: `ISO 8601`_ style timezone definition

        .. _ISO 8601: http://www.cl.cam.ac.uk/~mgk25/iso-time.html

        """
        super(TzOffset, self).__init__()
        hours, minutes = map(int, tzstring.split(':'))

        self.__offset = datetime.timedelta(hours=hours, minutes=minutes)

    def __repr__(self):
        """Self-documenting string representation.

        :rtype: ``str``
        :return: String to recreate ``TzOffset`` object

        """
        return repr_assist(self, {'tzstring': self.as_timezone()})

    def dst(self, dt=None):
        """Daylight Savings Time offset.

        .. note::
           This method is only for compatibility with the ``tzinfo`` interface,
           and does nothing

        :param dt: For compatibility with parent classes

        """
        return datetime.timedelta(0)

    def as_timezone(self):
        """Create a human-readable timezone string.

        :rtype: ``str``
        :return: Human-readable timezone definition

        """
        offset = self.utcoffset()
        hours, minutes = divmod(offset.seconds / 60, 60)
        if offset.days == -1:
            hours = -24 + hours

        return '%+03i:%02i' % (hours, minutes)

    def utcoffset(self, dt=None):
        """Return the offset in minutes from UTC.

        :param dt: For compatibility with parent classes

        """
        return self.__offset


class Timestamp(datetime.datetime):

    """Class for representing an OSM timestamp value."""

    def isoformat(self):
        """Generate an ISO 8601 formatted time stamp.

        :rtype: ``str``
        :return: `ISO 8601`_ formatted time stamp

        .. _ISO 8601: http://www.cl.cam.ac.uk/~mgk25/iso-time.html

        """
        text = [self.strftime('%Y-%m-%dT%H:%M:%S'), ]
        if self.tzinfo:
            text.append(self.tzinfo.as_timezone())
        else:
            text.append('+00:00')
        return ''.join(text)

    @staticmethod
    def parse_isoformat(timestamp):
        """Parse an ISO 8601 formatted time stamp.

        :param str timestamp: Timestamp to parse
        :rtype: ``Timestamp``
        :return: Parsed timestamp

        """
        if len(timestamp) == 20:
            zone = TzOffset('+00:00')
            timestamp = timestamp[:-1]
        elif len(timestamp) == 24:
            zone = TzOffset('%s:%s' % (timestamp[-5:-2], timestamp[-2:]))
            timestamp = timestamp[:-5]
        elif len(timestamp) == 25:
            zone = TzOffset(timestamp[-6:])
            timestamp = timestamp[:-6]
        timestamp = Timestamp.strptime(timestamp, '%Y-%m-%dT%H:%M:%S')
        timestamp = timestamp.replace(tzinfo=zone)
        return timestamp

#}

#{ Coordinate conversion utilities

iso6709_matcher = re.compile(r'^([-+][\d.]+)([-+][\d.]+)([+-][\d.]+)?/$')


def from_iso6709(coordinates):
    """Parse ISO 6709 coordinate strings.

    This function will parse ISO 6709-1983(E) "Standard representation of
    latitude, longitude and altitude for geographic point locations" elements.
    Unfortunately, the standard is rather convoluted and this implementation is
    incomplete, but it does support most of the common formats in the wild.

    The W3C has a simplified profile for ISO 6709 in `Latitude, Longitude and
    Altitude format for geospatial information`_.  It unfortunately hasn't
    received widespread support as yet, but hopefully it will grow just as the
    `simplified ISO 8601 profile`_ has.

    .. seealso::

       :func:`to_iso6709`

    :page str coordinates: ISO 6709 coordinates string
    :rtype: ``tuple``
    :return: A tuple consisting of latitude and longitude in degrees, along
        with the elevation in metres
    :raise ValueError: Input string is not ISO 6709 compliant
    :raise ValueError: Invalid value for latitude
    :raise ValueError: Invalid value for longitude

    .. _simplified ISO 8601 profile: http://www.w3.org/TR/NOTE-datetime

    """
    matches = iso6709_matcher.match(coordinates)
    if matches:
        latitude, longitude, altitude = matches.groups()
    else:
        raise ValueError('Incorrect format for string')
    sign = 1 if latitude[0] == '+' else -1
    latitude_head = len(latitude.split('.')[0])
    if latitude_head == 3:  # ±DD(.D{1,4})?
        latitude = float(latitude)
    elif latitude_head == 5:  # ±DDMM(.M{1,4})?
        latitude = float(latitude[:3]) + (sign * (float(latitude[3:]) / 60))
    elif latitude_head == 7:  # ±DDMMSS(.S{1,4})?
        latitude = float(latitude[:3]) + (sign * (float(latitude[3:5]) / 60)) \
            + (sign * (float(latitude[5:]) / 3600))
    else:
        raise ValueError('Incorrect format for latitude %r' % latitude)
    sign = 1 if longitude[0] == '+' else -1
    longitude_head = len(longitude.split('.')[0])
    if longitude_head == 4:  # ±DDD(.D{1,4})?
        longitude = float(longitude)
    elif longitude_head == 6:  # ±DDDMM(.M{1,4})?
        longitude = float(longitude[:4]) + (sign * (float(longitude[4:]) / 60))
    elif longitude_head == 8:  # ±DDDMMSS(.S{1,4})?
        longitude = float(longitude[:4]) \
            + (sign * (float(longitude[4:6]) / 60)) \
            + (sign * (float(longitude[6:]) / 3600))
    else:
        raise ValueError('Incorrect format for longitude %r' % longitude)
    if altitude:
        altitude = float(altitude)
    return latitude, longitude, altitude


def to_iso6709(latitude, longitude, altitude=None, format='dd', precision=4):
    """Produce ISO 6709 coordinate strings.

    This function will produce ISO 6709-1983(E) "Standard representation of
    latitude, longitude and altitude for geographic point locations" elements.

    .. seealso::

       :func:`from_iso6709`

    :param float latitude: Location's latitude
    :param float longitude: Location's longitude
    :param float altitude: Location's altitude
    :param str format: Format type for string
    :param int precision: Latitude/longitude precision
    :rtype: ``str``
    :return: ISO 6709 coordinates string
    :raise ValueError: Unknown value for ``format``

    .. _Latitude, Longitude and Altitude format for geospatial information:
       http://www.w3.org/2005/Incubator/geo/Wiki/LatitudeLongitudeAltitude
    .. _wikipedia ISO 6709 page: http://en.wikipedia.org/wiki/ISO_6709

    """
    text = []
    if format == 'd':
        text.append('%+03i%+04i' % (latitude, longitude))
    elif format == 'dd':
        text.append('%+0*.*f%+0*.*f' % (precision + 4, precision, latitude,
                                        precision + 5, precision, longitude))
    elif format in ('dm', 'dms'):
        if format == 'dm':
            latitude_dms = to_dms(latitude, 'dm')
            longitude_dms = to_dms(longitude, 'dm')
        elif format == 'dms':
            latitude_dms = to_dms(latitude)
            longitude_dms = to_dms(longitude)
        latitude_sign = '-' if any(i < 0 for i in latitude_dms) else '+'
        latitude_dms = tuple(abs(i) for i in latitude_dms)
        longitude_sign = '-' if any(i < 0 for i in longitude_dms) else '+'
        longitude_dms = tuple(abs(i) for i in longitude_dms)
        if format == 'dm':
            text.append('%s%02i%02i' % ((latitude_sign, ) + latitude_dms))
            text.append('%s%03i%02i' % ((longitude_sign, ) + longitude_dms))
        elif format == 'dms':
            text.append('%s%02i%02i%02i' % ((latitude_sign, ) + latitude_dms))
            text.append('%s%03i%02i%02i'
                        % ((longitude_sign, ) + longitude_dms))
    else:
        raise ValueError('Unknown format type %r' % format)
    if altitude and int(altitude) == altitude:
        text.append('%+i' % altitude)
    elif altitude:
        text.append('%+.3f' % altitude)
    text.append('/')
    return ''.join(text)


def angle_to_distance(angle, units='metric'):
    """Convert angle in to distance along a great circle.

    :param float angle: Angle in degrees to convert to distance
    :param str units: Unit type to be used for distances
    :rtype: ``float``
    :return: Distance in ``units``
    :raise ValueError: Unknown value for ``units``

    """
    distance = math.radians(angle) * BODY_RADIUS

    if units in ('km', 'metric'):
        return distance
    elif units in ('sm', 'imperial', 'US customary'):
        return distance / STATUTE_MILE
    elif units in ('nm', 'nautical'):
        return distance / NAUTICAL_MILE
    else:
        raise ValueError('Unknown units type %r' % units)


def distance_to_angle(distance, units='metric'):
    """Convert a distance in to an angle along a great circle.

    :param float distance: Distance to convert to degrees
    :param str units: Unit type to be used for distances
    :rtype: ``float``
    :return: Angle in degrees
    :raise ValueError: Unknown value for ``units``

    """
    if units in ('km', 'metric'):
        pass
    elif units in ('sm', 'imperial', 'US customary'):
        distance *= STATUTE_MILE
    elif units in ('nm', 'nautical'):
        distance *= NAUTICAL_MILE
    else:
        raise ValueError('Unknown units type %r' % units)

    return math.degrees(distance / BODY_RADIUS)


def from_grid_locator(locator):
    """Calculate geodesic latitude/longitude from Maidenhead locator.

    :param str locator: Maidenhead locator string
    :rtype: ``tuple`` of ``float`` objects
    :return: Geodesic latitude and longitude values
    :raise ValueError: Incorrect grid locator length
    :raise ValueError: Invalid values in locator string

    """
    if not len(locator) in (4, 6, 8):
        raise ValueError('Locator must be 4, 6 or 8 characters long %r'
                         % locator)

    # Convert the locator string to a list, because we need it to be mutable to
    # munge the values
    locator = list(locator)

    # Convert characters to numeric value, fields are always uppercase
    locator[0] = ord(locator[0]) - 65
    locator[1] = ord(locator[1]) - 65

    # Values for square are always integers
    locator[2] = int(locator[2])
    locator[3] = int(locator[3])

    if len(locator) >= 6:
        # Some people use uppercase for the subsquare data, in spite of
        # lowercase being the accepted style, so handle that too.
        locator[4] = ord(locator[4].lower()) - 97
        locator[5] = ord(locator[5].lower()) - 97

    if len(locator) == 8:
        # Extended square values are always integers
        locator[6] = int(locator[6])
        locator[7] = int(locator[7])

    # Check field values within 'A'(0) to 'R'(17), and square values are within
    # 0 to 9
    if not 0 <= locator[0] <= 17 \
       or not 0 <= locator[1] <= 17 \
       or not 0 <= locator[2] <= 9 \
       or not 0 <= locator[3] <= 9:
        raise ValueError('Invalid values in locator %r' % locator)

    # Check subsquare values are within 'a'(0) to 'x'(23)
    if len(locator) >= 6:
        if not 0 <= locator[4] <= 23 \
           or not 0 <= locator[5] <= 23:
            raise ValueError('Invalid values in locator %r' % locator)

    # Extended square values must be within 0 to 9
    if len(locator) == 8:
        if not 0 <= locator[6] <= 9 \
           or not 0 <= locator[7] <= 9:
            raise ValueError('Invalid values in locator %r' % locator)

    longitude = LONGITUDE_FIELD * locator[0] \
        + LONGITUDE_SQUARE * locator[2]
    latitude = LATITUDE_FIELD * locator[1] \
        + LATITUDE_SQUARE * locator[3]

    if len(locator) >= 6:
        longitude += LONGITUDE_SUBSQUARE * locator[4]
        latitude += LATITUDE_SUBSQUARE * locator[5]

    if len(locator) == 8:
        longitude += LONGITUDE_EXTSQUARE * locator[6] + LONGITUDE_EXTSQUARE / 2
        latitude += LATITUDE_EXTSQUARE * locator[7] + LATITUDE_EXTSQUARE / 2
    else:
        longitude += LONGITUDE_EXTSQUARE * 5
        latitude += LATITUDE_EXTSQUARE * 5

    # Rebase longitude and latitude to normal geodesic
    longitude -= 180
    latitude -= 90

    return latitude, longitude


def to_grid_locator(latitude, longitude, precision='square'):
    """Calculate Maidenhead locator from latitude and longitude.

    :param float latitude: Position's latitude
    :param float longitude: Position's longitude
    :param str precision: Precision with which generate locator string
    :rtype: ``str``
    :return: Maidenhead locator for latitude and longitude
    :raise ValueError: Invalid precision identifier
    :raise ValueError: Invalid latitude or longitude value

    """
    if not precision in ('square', 'subsquare', 'extsquare'):
        raise ValueError('Unsupported precision value %r' % precision)

    if not -90 <= latitude <= 90:
        raise ValueError('Invalid latitude value %r' % latitude)
    if not -180 <= longitude <= 180:
        raise ValueError('Invalid longitude value %r' % longitude)

    latitude += 90.0
    longitude += 180.0

    locator = []

    field = int(longitude / LONGITUDE_FIELD)
    locator.append(chr(field + 65))
    longitude -= field * LONGITUDE_FIELD

    field = int(latitude / LATITUDE_FIELD)
    locator.append(chr(field + 65))
    latitude -= field * LATITUDE_FIELD

    square = int(longitude / LONGITUDE_SQUARE)
    locator.append(str(square))
    longitude -= square * LONGITUDE_SQUARE

    square = int(latitude / LATITUDE_SQUARE)
    locator.append(str(square))
    latitude -= square * LATITUDE_SQUARE

    if precision in ('subsquare', 'extsquare'):
        subsquare = int(longitude / LONGITUDE_SUBSQUARE)
        locator.append(chr(subsquare + 97))
        longitude -= subsquare * LONGITUDE_SUBSQUARE

        subsquare = int(latitude / LATITUDE_SUBSQUARE)
        locator.append(chr(subsquare + 97))
        latitude -= subsquare * LATITUDE_SUBSQUARE

    if precision == 'extsquare':
        extsquare = int(longitude / LONGITUDE_EXTSQUARE)
        locator.append(str(extsquare))

        extsquare = int(latitude / LATITUDE_EXTSQUARE)
        locator.append(str(extsquare))

    return ''.join(locator)


def parse_location(location):
    """Parse latitude and longitude from string location.

    :param str location: String to parse
    :rtype: ``tuple`` of ``float`` objects
    :return: Latitude and longitude of location

    """

    def split_dms(text, hemisphere):
        """Split degrees, minutes and seconds string.

        :param str text: Text to split
        :rtype: ``float``
        :return: Decimal degrees

        """
        out = []
        sect = []
        for i in text:
            if i.isdigit():
                sect.append(i)
            else:
                out.append(sect)
                sect = []
        d, m, s = [float(''.join(i)) for i in out]
        if hemisphere in 'SW':
            d, m, s = [-1 * x for x in (d, m, s)]
        return to_dd(d, m, s)

    for sep in ';, ':
        chunks = location.split(sep)
        if len(chunks) == 2:
            if chunks[0].endswith('N'):
                latitude = float(chunks[0][:-1])
            elif chunks[0].endswith('S'):
                latitude = -1 * float(chunks[0][:-1])
            else:
                latitude = float(chunks[0])
            if chunks[1].endswith('E'):
                longitude = float(chunks[1][:-1])
            elif chunks[1].endswith('W'):
                longitude = -1 * float(chunks[1][:-1])
            else:
                longitude = float(chunks[1])
            return latitude, longitude
        elif len(chunks) == 4:
            if chunks[0].endswith(('s', '"')):
                latitude = split_dms(chunks[0], chunks[1])
            else:
                latitude = float(chunks[0])
                if chunks[1] == 'S':
                    latitude = -1 * latitude
            if chunks[2].endswith(('s', '"')):
                longitude = split_dms(chunks[2], chunks[3])
            else:
                longitude = float(chunks[2])
                if chunks[3] == 'W':
                    longitude = -1 * longitude
            return latitude, longitude
#}

#{ Solar event utilities

#: Sunrise/-set mappings from name to angle
ZENITH = {
    # All values are specified in degrees!

    # Sunrise/sunset is defined as the moment the upper limb becomes visible,
    # taking in to account atmospheric refraction.  That is 34' for atmospheric
    # refraction and 16' for angle between the Sun's centre and it's upper
    # limb, resulting in a combined 50' below the horizon.
    #
    # We're totally ignoring how temperature and pressure change the amount of
    # atmospheric refraction, because their effects are drowned out by rounding
    # errors in the equation.
    None: -50 / 60,

    # Twilight definitions specify the angle in degrees of the Sun below the
    # horizon
    'civil': -6,
    'nautical': -12,
    'astronomical': -18,
}


def sun_rise_set(latitude, longitude, date, mode='rise', timezone=0,
                 zenith=None):
    """Calculate sunrise or sunset for a specific location.

    This function calculates the time sunrise or sunset, or optionally the
    beginning or end of a specified twilight period.

    Source::

        Almanac for Computers, 1990
        published by Nautical Almanac Office
        United States Naval Observatory
        Washington, DC 20392

    :param float latitude: Location's latitude
    :param float longitude: Location's longitude
    :param datetime.date date: Calculate rise or set for given date
    :param str mode: Which time to calculate
    :param int timezone: Offset from UTC in minutes
    :param str zenith: Calculate rise/set events, or twilight times
    :rtype: :class:`datetime.time` or ``None``
    :return: The time for the given event in the specified timezone, or
        ``None`` if the event doesn't occur on the given date
    :raise ValueError: Unknown value for ``mode``

    """
    if not date:
        date = datetime.date.today()

    zenith = ZENITH[zenith]

    # First calculate the day of the year
    # Thanks, datetime this would have been ugly without you!!!
    n = (date - datetime.date(date.year - 1, 12, 31)).days

    # Convert the longitude to hour value and calculate an approximate time
    lng_hour = longitude / 15

    if mode == 'rise':
        t = n + ((6 - lng_hour) / 24)
    elif mode == 'set':
        t = n + ((18 - lng_hour) / 24)
    else:
        raise ValueError('Unknown mode value %r' % mode)

    # Calculate the Sun's mean anomaly
    m = (0.9856 * t) - 3.289

    # Calculate the Sun's true longitude
    l = m + 1.916 * math.sin(math.radians(m)) + 0.020 \
        * math.sin(2 * math.radians(m)) + 282.634
    l = abs(l) % 360

    # Calculate the Sun's right ascension
    ra = math.degrees(math.atan(0.91764 * math.tan(math.radians(l))))

    # Right ascension value needs to be in the same quadrant as L
    l_quandrant = (math.floor(l / 90)) * 90
    ra_quandrant = (math.floor(ra / 90)) * 90
    ra = ra + (l_quandrant - ra_quandrant)

    # Right ascension value needs to be converted into hours
    ra = ra / 15

    # Calculate the Sun's declination
    sin_dec = 0.39782 * math.sin(math.radians(l))
    cos_dec = math.cos(math.asin(sin_dec))

    # Calculate the Sun's local hour angle
    cos_h = (math.radians(zenith) -
             (sin_dec * math.sin(math.radians(latitude)))) \
        / (cos_dec * math.cos(math.radians(latitude)))

    if cos_h > 1:
        # The sun never rises on this location (on the specified date)
        return None
    elif cos_h < -1:
        # The sun never sets on this location (on the specified date)
        return None

    # Finish calculating H and convert into hours
    if mode == 'rise':
        h = 360 - math.degrees(math.acos(cos_h))
    else:
        h = math.degrees(math.acos(cos_h))
    h = h / 15

    # Calculate local mean time of rising/setting
    t = h + ra - (0.06571 * t) - 6.622

    # Adjust back to UTC
    utc = t - lng_hour

    # Convert UT value to local time zone of latitude/longitude
    local_t = utc + timezone / 60
    if local_t < 0:
        local_t += 24
    elif local_t > 23:
        local_t -= 24

    hour = int(local_t)
    if hour == 0:
        minute = int(60 * local_t)
    else:
        minute = int(60 * (local_t % hour))
    if minute < 0:
        minute += 60
    return datetime.time(hour, minute)


def sun_events(latitude, longitude, date, timezone=0, zenith=None):
    """Convenience function for calculating sunrise and sunset.

    Civil twilight starts/ends when the Sun's centre is 6 degrees below
    the horizon.

    Nautical twilight starts/ends when the Sun's centre is 12 degrees
    below the horizon.

    Astronomical twilight starts/ends when the Sun's centre is 18 degrees below
    the horizon.

    :param float latitude: Location's latitude
    :param float longitude: Location's longitude
    :param datetime.date date: Calculate rise or set for given date
    :param int timezone: Offset from UTC in minutes
    :param str zenith: Calculate rise/set events, or twilight times
    :rtype: ``tuple`` of :class:`datetime.time`
    :return: The time for the given events in the specified timezone

    """
    return (sun_rise_set(latitude, longitude, date, 'rise', timezone, zenith),
            sun_rise_set(latitude, longitude, date, 'set', timezone, zenith))

#}


def dump_xearth_markers(markers, name='identifier'):
    """Generate an Xearth compatible marker file.

    ``dump_xearth_markers()`` writes a simple Xearth_ marker file from
    a dictionary of :class:`trigpoints.Trigpoint` objects.

    It expects a dictionary in one of the following formats. For support of
    :class:`Trigpoint` that is::

        {500936: Trigpoint(52.066035, -0.281449, 37.0, "Broom Farm"),
         501097: Trigpoint(52.010585, -0.173443, 97.0, "Bygrave"),
         505392: Trigpoint(51.910886, -0.186462, 136.0, "Sish Lane")}

    And generates output of the form::

        52.066035 -0.281449 "500936" # Broom Farm, alt 37m
        52.010585 -0.173443 "501097" # Bygrave, alt 97m
        51.910886 -0.186462 "205392" # Sish Lane, alt 136m

    Or similar to the following if the ``name`` parameter is set to ``name``::

        52.066035 -0.281449 "Broom Farm" # 500936 alt 37m
        52.010585 -0.173443 "Bygrave" # 501097 alt 97m
        51.910886 -0.186462 "Sish Lane" # 205392 alt 136m

    Point objects should be provided in the following format::

        {"Broom Farm": Point(52.066035, -0.281449),
         "Bygrave": Point(52.010585, -0.173443),
         "Sish Lane": Point(51.910886, -0.186462)}

    And generates output of the form::

        52.066035 -0.281449 "Broom Farm"
        52.010585 -0.173443 "Bygrave"
        51.910886 -0.186462 "Sish Lane"

    .. note:
       xplanet_ also supports xearth marker files, and as such can use the
       output from this function.

    .. seealso:

       :func:`upoints.xearth.Xearths.import_locations`

    :param dict markers: Dictionary of identifier keys, with :class:`Trigpoint`
        values
    :param str name: Value to use as Xearth display string
    :rtype: ``list``
    :return: List of strings representing an Xearth marker file
    :raise ValueError: Unsupported value for ``name``

    .. _xearth: http://hewgill.com/xearth/original/
    .. _xplanet: http://xplanet.sourceforge.net/

    """
    output = []
    for identifier, point in markers.items():
        line = ['%f %f ' % (point.latitude, point.longitude), ]
        if hasattr(point, 'name') and point.name:
            if name == 'identifier':
                line.append('"%s" # %s' % (identifier, point.name))
            elif name == 'name':
                line.append('"%s" # %s' % (point.name, identifier))
            elif name == 'comment':
                line.append('"%s" # %s' % (identifier, point.comment))
            else:
                raise ValueError('Unknown name type %r' % name)
            if hasattr(point, 'altitude') and point.altitude:
                line.append(', alt %im' % point.altitude)
        else:
            line.append('"%s"' % identifier)
        output.append(''.join(line))
    # Return the list sorted on the marker name
    return sorted(output, key=lambda x: x.split()[2])


def calc_radius(latitude, ellipsoid='WGS84'):
    """Calculate earth radius for a given latitude.

    This function is most useful when dealing with datasets that are very
    localised and require the accuracy of an ellipsoid model without the
    complexity of code necessary to actually use one.  The results are meant to
    be used as a :data:`BODY_RADIUS` replacement when the simple geocentric
    value is not good enough.

    The original use for ``calc_radius`` is to set a more accurate radius value
    for use with trigpointing databases that are keyed on the OSGB36 datum, but
    it has been expanded to cover other ellipsoids.

    :param float latitude: Latitude to calculate earth radius for
    :type ellipsoid: ``tuple`` of ``float`` objects
    :param ellipsoid: Ellipsoid model to use for calculation
    :rtype: ``float``
    :return: Approximated Earth radius at the given latitude

    """

    ellipsoids = {
        'Airy (1830)': (6377.563, 6356.257),  # Ordnance Survey default
        'Bessel': (6377.397, 6356.079),
        'Clarke (1880)': (6378.249145, 6356.51486955),
        'FAI sphere': (6371, 6371),  # Idealised
        'GRS-67': (6378.160, 6356.775),
        'International': (6378.388, 6356.912),
        'Krasovsky': (6378.245, 6356.863),
        'NAD27': (6378.206, 6356.584),
        'WGS66': (6378.145, 6356.758),
        'WGS72': (6378.135, 6356.751),
        'WGS84': (6378.137, 6356.752),  # GPS default
    }

    # Equatorial radius, polar radius
    major, minor = ellipsoids[ellipsoid]
    # eccentricity of the ellipsoid
    eccentricity = 1 - (minor ** 2 / major ** 2)

    sl = math.sin(math.radians(latitude))
    return (major * (1 - eccentricity)) / (1 - eccentricity * sl ** 2) ** 1.5



#! /usr/bin/python -tt
# coding=utf-8
"""grab_net_sources - Fetch sources for tests"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import bz2
import gzip
import os
import sys
import tempfile
try:
    from urllib.parse import urlparse
    from urllib.request import (urlopen, urlretrieve)
except ImportError:
    from urlparse import urlparse
    from urllib import (urlopen, urlretrieve)


SOURCES = [
    'http://cvs.savannah.gnu.org/viewvc/*checkout*/miscfiles/cities.dat?root=miscfiles',
    'http://weather.noaa.gov/data/nsd_bbsss.txt',
    'http://weather.noaa.gov/data/nsd_cccc.txt',
    'http://www.haroldstreet.org.uk/download.php?file=waypoints/files/alltrigs-wgs84.txt',
    'http://myapp.fr/cellsIdData/cells.txt.gz',
    'http://xplanet.sourceforge.net/Extras/earth-markers-schaumann',
]

def data_file(resource):
    """Generate a local filename for the resource.

    >>> print data_file(SOURCES[0])
    tests/data/cities.dat
    >>> print data_file(SOURCES[4])
    tests/data/cells.txt
    >>> print data_file(SOURCES[3])
    tests/data/alltrigs-wgs84.txt

    :parameters str resource: Source
    :rtype: `str`
    :return: Local filename

    """
    filename = os.path.join(os.path.dirname(__file__), 'data',
                            os.path.basename(urlparse(resource[0]).path))
    if filename.endswith('.gz'):
        return filename[:-3]
    elif filename.endswith('.bz2'):
        return filename[:-4]
    else:
        return filename

def main(argv=None):
    """Main script handler.

    :param list argv: Command line arguments

    """
    print('*WARNING* This script will fetch some data files that can not be '
          'distributed legally!  In some jurisdictions you may not even be '
          'entitled to personal use of the data it fetches without express '
          'consent of the copyright holders.')
    if not argv:
        argv = sys.argv
    if len(argv) == 2 and argv[1] in ('-f' or '--force'):
        force = True
    else:
        force = False
    cached = 0
    for path in SOURCES:
        filename = data_file(path)
        if not force and os.path.exists(filename):
            print('%r already downloaded.' % path)
            cached += 1
        else:
            print('Fetching %r...' % path)
            if path.endswith('.gz'):
                temp = tempfile.mkstemp()[1]
                try:
                    urlretrieve(path, temp)
                    data = gzip.GzipFile(temp).read()
                finally:
                    os.unlink(temp)
                open(filename, 'w').write(data)
            elif path.endswith('.bz2'):
                data = bz2.decompress(urlopen(path).read())
                open(filename, 'w').write(data)
            else:
                urlretrieve(path, filename)
    if cached > 1:
        print("You can force download with the `-f' option to this script.")

if __name__ == '__main__':
    main(sys.argv)
